# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import numpy
import pytest
import skhep_testdata

import uproot

pytest.importorskip("awkward")


@pytest.mark.parametrize("is_forth", [False, True])
def test_00(is_forth):
    with uproot.open(skhep_testdata.data_path("issue367b.root")) as file:
        branch = file["tree/weights"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0]["0"][0] == "expskin_FluxUnisim"
        # py[-1] == <STLMap {'expskin_FluxUnisim': [0.944759093019904, 1.0890682745548674, ..., 1.1035170311451232, 0.8873957186284592], ...} at 0x7fbc4c1325e0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_01(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-delphes-pr442.root")) as file:
        branch = file["Delphes/GenJet/GenJet.SoftDroppedSubJet1"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][0]["fE"] == pytest.approx(84.56447925448748)
        assert py[0][0]["fP"]["fZ"] == pytest.approx(-81.600465)
        # py[-1] == array([<TLorentzVector (version 4) at 0x7fac4bb8d2b0>, <TLorentzVector (version 4) at 0x7fac4bb8d2e0>], dtype=object)
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_02(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-delphes-pr442.root")) as file:
        branch = file["Delphes/GenJet/GenJet.TrimmedP4[5]"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak")
        assert py[0][0][0]["fE"] == 0
        assert py[-1][8][3]["fE"] == 0
        # py[-1] == array([[<TLorentzVector (version 4) at 0x7fbfea1fa3d0>, <TLorentzVector (version 4) at 0x7fbfea1fa370>, <TLorentzVector (version 4) at 0x7fbfea1fa310>, <TLorentzVector (version 4) at 0x7fbfea1fa2b0>, <TLorentzVector (version 4) at 0x7fbfea1fa250>], [<TLorentzVector (version 4) at 0x7fbfea1fa1f0>, <TLorentzVector (version 4) at 0x7fbfea1fa190>, <TLorentzVector (version 4) at 0x7fbfea1fa130>, <TLorentzVector (version 4) at 0x7fbfea1fa0d0>, <TLorentzVector (version 4) at 0x7fbfea1fa070>]], dtype=object)
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_03(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-FCCDelphesOutput.root")) as file:
        branch = file["metadata/gaudiConfigOptions"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][0] == 'Converter.hepmcStatusList = "[]";\n'
        assert (
            py[0][14]
            == 'ToolSvc.photons.delphesArrayName = "PhotonEfficiency/photons";\nToolSvc.photons.isolationTags = "photonITags";\nToolSvc.photons.mcAssociations = "photonsToMC";\nToolSvc.photons.particles = "photons";\n'
        )
        # py[-1] == <STLVector ['Converter.hepmcStatusList = "[]";\n', ...] at 0x7f7fc23ca2e0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_04(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue-123a.root")) as file:
        branch = file["POOLContainerForm/DataHeaderForm/m_uints"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][0][0] == 2
        assert py[0][3][2] == 589824
        # py[-1] == <STLVector [[1], ...] at 0x7fc153ed6670>
        assert py.layout.form == interp.awkward_form(branch.file)


def test_05():
    with uproot.open(skhep_testdata.data_path("uproot-issue-123a.root")) as file:
        branch = file["CollectionTree/TrigConfKeys"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        py = branch.array(interp, library="np", entry_stop=2)
        # py[-1] == <xAOD::TrigConfKeys_v1 (version 1) at 0x7fecf9212760>


@pytest.mark.parametrize("is_forth", [False, True])
def test_06(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue-123a.root")) as file:
        branch = file["CollectionTree/AnalysisJetsAuxDyn.NumTrkPt500"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1][-1][-2] == 0
        # py[-1] == <STLVector [[7, 2, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], ...] at 0x7f1377d56610>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_07(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue-123a.root")) as file:
        branch = file["CollectionTree/AnalysisJetsAuxDyn.SumPtTrkPt500"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1][2][0] == pytest.approx(53949.015625)
        # py[-1] == <STLVector [[28132.615, 38298.348, 0.0, 0.0, 0.0, 665.71674, ..., 0.0, 0.0, 0.0, 0.0, 0.0, 0.0], ...] at 0x7f225f7496a0>
        assert py.layout.form == interp.awkward_form(branch.file)


def test_08():
    with uproot.open(skhep_testdata.data_path("uproot-issue-123a.root")) as file:
        branch = file["CollectionTree/AnalysisJetsAuxDyn.GhostTrack"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        py = branch.array(interp, library="np", entry_stop=2)
        # py[-1] == <STLVector [[<ElementLink<DataVector<xAOD::IParticle>> (version 1) at 0x7fc6a08f2f70>, ...], ...] at 0x7fc6a08f2f10>


@pytest.mark.parametrize("is_forth", [False, True])
def test_09(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue-123a.root")) as file:
        branch = file["CollectionTree"][
            "AntiKt10UFOCSSKJetsAuxDyn.GhostVR30Rmax4Rmin02TrackJet_BTagging201903"
        ]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][0][0]["m_persKey"] == 352341021
        # py[-1] == <STLVector [[<ElementLink<DataVector<xAOD::IParticle>> (version 1) at 0x7febbf1b2fa0>], ...] at 0x7febbf1b2f40>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_10(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue-123a.root")) as file:
        branch = file["CollectionTree/CaloCalTopoClustersAuxDyn.e_sampl"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=3)
        assert py[0][0][0] == pytest.approx(168.03048706054688)
        # py[-1] == <STLVector [[4499.692, 26934.783, 75009.02, 676.8455, 15.539482], ...] at 0x7fe3240a09a0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_11(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue-123a.root")) as file:
        branch = file["CollectionTree"][
            "TruthBosonsWithDecayVerticesAuxDyn.incomingParticleLinks"
        ]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][0][0]["m_persKey"] == 921521854
        # py[-1] == <STLVector [[<ElementLink<DataVector<xAOD::TruthParticle_v1>> (version 1) at 0x7f636a9484c0>], ...] at 0x7f636a948eb0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_12(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue-123a.root")) as file:
        branch = file["CollectionTree/TruthBottomAuxDyn.parentLinks"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][0][0]["m_persIndex"] == 2
        # py[-1] == <STLVector [[<ElementLink<DataVector<xAOD::TruthParticle_v1>> (version 1) at 0x7fc259ae37c0>], ...] at 0x7fc259ae3f10>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_13(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue-123a.root")) as file:
        branch = file["CollectionTree/egammaClustersAuxDyn.constituentClusterLinks"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=3)
        assert py[0][0][0]["m_persIndex"] == 0
        # py[-1] == <STLVector [[<ElementLink<DataVector<xAOD::CaloCluster_v1>> (version 1) at 0x7fa94e968c10>], ...] at 0x7fa94e968c70>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_14(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue-123a.root")) as file:
        branch = file["CollectionTree/egammaClustersAuxDyn.eta_sampl"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=3)
        assert py[0][0][-1] == pytest.approx(0.4663555920124054)
        # py[-1] == <STLVector [[-0.53503126, -0.5374735, -0.5373216, -0.52376634]] at 0x7f5d2ba80d30>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_15(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue-123a.root")) as file:
        branch = file["CollectionTree/AnalysisHLT_mu24_ilooseAuxDyn.TrigMatchedObjects"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1][-1][-1]["m_persKey"] == 980095599
        # py[-1] == <STLVector [[<ElementLink<DataVector<xAOD::IParticle>> (version 1) at 0x7fb9b9d24c10>], ...] at 0x7fb9b9d29250>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_16(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue-123a.root")) as file:
        branch = file["CollectionTree/AnalysisHLT_mu40AuxDyn.TrigMatchedObjects"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1][0]["m_persKey"][0] == 980095599
        # py[-1] == <STLVector [[<ElementLink<DataVector<xAOD::IParticle>> (version 1) at 0x7f6e29be5cd0>], ...] at 0x7f6e29bea250>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_17(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue390.root")) as file:
        branch = file["E/Evt/AAObject/usr_names"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=None)
        assert len(py[0]) == 0
        # py[-1] == <STLVector [] at 0x7f6afa6ead00>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_18(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue390.root")) as file:
        branch = file["E/Evt/hits/hits.t"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][0] == pytest.approx(70104010.0)
        # py[-1] == array([80860738., 80861689., 80862014., 80861709., 80861737., 80861158., 80862362., 80860821., 80862271., 80862273., 80861294., 80861860., 80862548., 80861733., 80861605., 80860467., 80860408., 80861562., 80862012., 80862350., 80861491., 80860384., 80860930., 80861541., 80861461., 80861749., 80862352., 80861813., 80861822., 80861871., 80862000., 80862255., 80862253., 80862249., 80862266., 80862248., 80862246., 80862847., 80863032., 80861952., 80861954., 80861953., 80861957., 80861951., 80861961., 80861959., 80861955., 80861994., 80862060., 80861971., 80862004., 80862002., 80862059., 80861695., 80861813., 80861967., 80862919., 80862043., 80862054., 80862044., 80862044., 80862040., 80862043., 80862037., 80862040., 80862039., 80862070., 80862042., 80862322., 80861605., 80861865., 80863034., 80862987., 80861545., 80860392., 80861003., 80861564., 80862109., 80861821., 80862083., 80861121., 80862513., 80862513., 80862731., 80861604., 80862003., 80861910., 80861854., 80862297., 80860989., 80862948., 80862075., 80862141., 80862117., 80862039., 80862114., 80862075., 80862042., 80862072., 80862439., 80862481., 80861656., 80862096., 80862215., 80862215., 80862195., 80862458., 80862432., 80861915., 80861012., 80862208., 80861885., 80861888., 80861994., 80861883., 80862194., 80861812., 80862184., 80862309., 80862297., 80862840., 80862400., 80861565., 80862226., 80862149.])
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_19(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue390.root")) as file:
        branch = file["E/Evt/hits/hits.a"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][-1] == 0.0
        # py[-1] == array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.])
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_20(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue390.root")) as file:
        branch = file["E/Evt/hits/hits.trig"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][5] == 1
        # py[-1] == array([0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 1, 1, 1, 1, 0, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 0, 1, 1, 1, 1, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0], dtype=int32)
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_21(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue390.root")) as file:
        branch = file["E/Evt/hits/hits.tot"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][0] == 24
        # py[-1] == array([29, 26, 22, 18, 22, 28, 28, 28, 21, 24, 28, 30, 25, 24, 30, 28, 29, 4, 21, 25, 26, 22, 23, 22, 23, 29, 23, 30, 24, 29, 31, 27, 32, 28, 30, 33, 33, 31, 29, 18, 23, 34, 21, 33, 33, 29, 37, 23, 21, 40, 25, 29, 22, 17, 31, 25, 28, 26, 21, 20, 25, 51, 38, 64, 42, 28, 29, 26, 21, 31, 22, 18, 41, 28, 29, 28, 29, 15, 25, 27, 24, 28, 28, 34, 28, 21, 19, 21, 20, 24, 26, 24, 13, 22, 30, 25, 17, 27, 24, 16, 31, 27, 29, 23, 26, 25, 26, 28, 12, 18, 30, 27, 48, 16, 25, 24, 27, 10, 21, 25, 30, 26, 26, 28, 24], dtype=uint32)
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_22(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue390.root")) as file:
        branch = file["E/Evt/hits/hits.pos.x"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][2] == 0.0
        # py[-1] == array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.])
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_23(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue390.root")) as file:
        branch = file["E/Evt/trks/trks.usr_names"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert len(py[0][0]) == 0
        # py[-1] == array([<STLVector [] at 0x7f09da748f70>, <STLVector [] at 0x7f09da75a5b0>, <STLVector [] at 0x7f09da75a610>, <STLVector [] at 0x7f09da75a670>, <STLVector [] at 0x7f09da75a6d0>, <STLVector [] at 0x7f09da75a730>, <STLVector [] at 0x7f09da75a790>, <STLVector [] at 0x7f09da75a7f0>, <STLVector [] at 0x7f09da75a850>, <STLVector [] at 0x7f09da75a8b0>, <STLVector [] at 0x7f09da75a910>, <STLVector [] at 0x7f09da75a970>, <STLVector [] at 0x7f09da75a9d0>, <STLVector [] at 0x7f09da75aa30>, <STLVector [] at 0x7f09da75aa90>, <STLVector [] at 0x7f09da75aaf0>, <STLVector [] at 0x7f09da75ab50>, <STLVector [] at 0x7f09da75abb0>, <STLVector [] at 0x7f09da75ac10>, <STLVector [] at 0x7f09da75ac70>, <STLVector [] at 0x7f09da75acd0>, <STLVector [] at 0x7f09da75ad30>, <STLVector [] at 0x7f09da75ad90>, <STLVector [] at 0x7f09da75adf0>, <STLVector [] at 0x7f09da75ae50>, <STLVector [] at 0x7f09da75aeb0>, <STLVector [] at 0x7f09da75af10>, <STLVector [] at 0x7f09da75af70>, <STLVector [] at 0x7f09da75afd0>, <STLVector [] at 0x7f09da75f070>, <STLVector [] at 0x7f09da75f0d0>, <STLVector [] at 0x7f09da75f130>, <STLVector [] at 0x7f09da75f190>, <STLVector [] at 0x7f09da75f1f0>, <STLVector [] at 0x7f09da75f250>, <STLVector [] at 0x7f09da75f2b0>, <STLVector [] at 0x7f09da75f310>, <STLVector [] at 0x7f09da75f370>, <STLVector [] at 0x7f09da75f3d0>, <STLVector [] at 0x7f09da75f430>, <STLVector [] at 0x7f09da75f490>, <STLVector [] at 0x7f09da75f4f0>, <STLVector [] at 0x7f09da75f550>, <STLVector [] at 0x7f09da75f5b0>, <STLVector [] at 0x7f09da75f610>, <STLVector [] at 0x7f09da75f670>, <STLVector [] at 0x7f09da75f6d0>, <STLVector [] at 0x7f09da75f730>, <STLVector [] at 0x7f09da75f790>, <STLVector [] at 0x7f09da75f7f0>, <STLVector [] at 0x7f09da75f850>, <STLVector [] at 0x7f09da75f8b0>, <STLVector [] at 0x7f09da75f910>, <STLVector [] at 0x7f09da75f970>, <STLVector [] at 0x7f09da75f9d0>], dtype=object)
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_24(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue390.root")) as file:
        branch = file["E/Evt/trks/trks.rec_stages"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][1][-1] == 5
        # py[-1] == array([<STLVector [1, 3, 5, 4] at 0x7fea61938d60>, <STLVector [1, 3, 5] at 0x7fea6194a4f0>, <STLVector [1, 3] at 0x7fea6194a580>, <STLVector [1, 3] at 0x7fea6194a5e0>, <STLVector [1, 3] at 0x7fea6194a640>, <STLVector [1, 3] at 0x7fea6194a6a0>, <STLVector [1, 3] at 0x7fea6194a700>, <STLVector [1, 3] at 0x7fea6194a760>, <STLVector [1, 3] at 0x7fea6194a7c0>, <STLVector [1, 3] at 0x7fea6194a820>, <STLVector [1, 3] at 0x7fea6194a880>, <STLVector [1, 3] at 0x7fea6194a8e0>, <STLVector [1, 3] at 0x7fea6194a940>, <STLVector [1, 3] at 0x7fea6194a9a0>, <STLVector [1, 3] at 0x7fea6194aa00>, <STLVector [1, 3] at 0x7fea6194aa60>, <STLVector [1, 3] at 0x7fea6194aac0>, <STLVector [1, 3] at 0x7fea6194ab20>, <STLVector [1, 3] at 0x7fea6194ab80>, <STLVector [1] at 0x7fea6194abe0>, <STLVector [1] at 0x7fea6194ac40>, <STLVector [1] at 0x7fea6194aca0>, <STLVector [1] at 0x7fea6194ad00>, <STLVector [1] at 0x7fea6194ad60>, <STLVector [1] at 0x7fea6194adc0>, <STLVector [1] at 0x7fea6194ae20>, <STLVector [1] at 0x7fea6194ae80>, <STLVector [1] at 0x7fea6194aee0>, <STLVector [1] at 0x7fea6194af40>, <STLVector [1] at 0x7fea6194afa0>, <STLVector [1] at 0x7fea6194e040>, <STLVector [1] at 0x7fea6194e0a0>, <STLVector [1] at 0x7fea6194e100>, <STLVector [1] at 0x7fea6194e160>, <STLVector [1] at 0x7fea6194e1c0>, <STLVector [1] at 0x7fea6194e220>, <STLVector [1] at 0x7fea6194e280>, <STLVector [1] at 0x7fea6194e2e0>, <STLVector [1] at 0x7fea6194e340>, <STLVector [1] at 0x7fea6194e3a0>, <STLVector [1] at 0x7fea6194e400>, <STLVector [1] at 0x7fea6194e460>, <STLVector [1] at 0x7fea6194e4c0>, <STLVector [1] at 0x7fea6194e520>, <STLVector [1] at 0x7fea6194e580>, <STLVector [1] at 0x7fea6194e5e0>, <STLVector [1] at 0x7fea6194e640>, <STLVector [1] at 0x7fea6194e6a0>, <STLVector [1] at 0x7fea6194e700>, <STLVector [1] at 0x7fea6194e760>, <STLVector [1] at 0x7fea6194e7c0>, <STLVector [1] at 0x7fea6194e820>, <STLVector [1] at 0x7fea6194e880>, <STLVector [1] at 0x7fea6194e8e0>, <STLVector [1] at 0x7fea6194e940>], dtype=object)
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_25(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue-208.root")) as file:
        branch = file["config/VERSION/VERSION._name"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][-1] == "numuCCAnalysis"
        # py[-1] == array(['psychePolicy', 'psycheEventModel', 'psycheCore', 'psycheUtils', 'psycheND280Utils', 'psycheIO', 'psycheSelections', 'psycheSystematics', 'highlandEventModel', 'highlandTools', 'highlandCore', 'highlandCorrections', 'highlandIO', 'baseAnalysis', 'baseTrackerAnalysis', 'numuCCAnalysis'], dtype=object)
        assert py.layout.form == interp.awkward_form(branch.file)


def test_26():
    with uproot.open(skhep_testdata.data_path("uproot-issue-208.root")) as file:
        branch = file["config/SEL/SEL._firstSteps"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        py = branch.array(interp, library="np", entry_stop=2)
        # py[-1] == array([<STLVector [] at 0x7f5060ef49d0>], dtype=object)


@pytest.mark.parametrize("is_forth", [False, True])
def test_27(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue-208.root")) as file:
        branch = file["config/SEL/SEL._branchAlias"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][0][-1] == "muFGD+Np"
        # py[-1] == array([<STLVector ['muTPC', 'muTPC+pTPC', ..., 'muFGD+Np'] at 0x7f1d79005a90>], dtype=object)
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_28(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue-208.root")) as file:
        branch = file["config/SEL/SEL._nCutsInBranch"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][0][-1] == 10
        # py[-1] == array([<STLVector [8, 9, 10, 11, 10, 7, 6, 7, 7, 8, 11, 10] at 0x7f409846fa00>], dtype=object)
        assert py.layout.form == interp.awkward_form(branch.file)


def test_29():
    with uproot.open(skhep_testdata.data_path("uproot-issue213.root")) as file:
        branch = file["T/eventPack/fGenInfo"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        py = branch.array(interp, library="np", entry_stop=2)
        # py[-1] == <JPetGeantEventInformation (version 3) at 0x7fd5bdeedac0>


@pytest.mark.parametrize("is_forth", [False, True])
def test_30(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue243-new.root")) as file:
        branch = file["triggerList/triggerMap/triggerMap.first"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=16)
        assert len(py[0]) == 0
        # py[-1] == array(['HLT_2j35_bmv2c2060_split_2j35_L14J15.0ETA25', 'HLT_j100_2j55_bmv2c2060_split'], dtype=object)
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_31(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue-308.root")) as file:
        branch = file["MetaData/BranchIDLists"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][0][-1] == 2428801822
        # py[-1] == <STLVector [[1971320715, 1805338087, 475485005, ..., 2417357619, 128868952, 2428801822], ...] at 0x7f0a7dd4a1f0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_32(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue31.root")) as file:
        branch = file["T/data/name"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[-1] == "two"
        # py[-1] == "two"
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_33(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue367b.root")) as file:
        branch = file["tree/weights"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][4]["1"][-1] == 1.0
        # py[-1] == <STLMap {'expskin_FluxUnisim': [0.944759093019904, 1.0890682745548674, ..., 1.1035170311451232, 0.8873957186284592], ...} at 0x7f4443068ca0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_34(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue371.root")) as file:
        branch = file["Header/Header./Header.geant4Version"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[-1] == "$Name: geant4-10-05-patch-01 $"
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_35(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue371.root")) as file:
        branch = file["Geant4Data/Geant4Data./Geant4Data.particles"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0, "1", 0, "name"] == "anti_alpha"
        # py[-1] == <STLMap {-1000020040: <BDSOutputROOTGeant4Data::ParticleInfo (version 1) at 0x7fb557996df0>, ...} at 0x7fb557a012e0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_36(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue371.root")) as file:
        branch = file["Model/Model./Model.samplerNamesUnique"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][-1] == "MP_F_99."
        # py[-1] == <STLVector ['DRIFT_0.', 'PRXSE01A.', ..., 'PRBHF_99.', 'MP_F_99.'] at 0x7f22f206df10>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_37(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue371.root")) as file:
        branch = file["Model/Model./Model.staPos"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][555]["fZ"] == 100.94856572890848
        # py[-1] == array([<TVector3 (version 3) at 0x7f3385c9cbe0>, <TVector3 (version 3) at 0x7f3385c9cc10>, <TVector3 (version 3) at 0x7f3385c9cc40>, ..., <TVector3 (version 3) at 0x7f3385b0cfd0>, <TVector3 (version 3) at 0x7f3385a9e040>, <TVector3 (version 3) at 0x7f3385a9e070>], dtype=object)
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_38(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue371.root")) as file:
        branch = file["Event/PRBHF_46."]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0]["zp"][0] == pytest.approx(0.9999998807907104)
        # py[-1] == <BDSOutputROOTEventSampler<float> (version 4) at 0x7f9be2f7b2e0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_39(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue399.root")) as file:
        branch = file["Geant4Data/Geant4Data./Geant4Data.particles"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[-1][1]["0"] == -1000020030
        # py[-1] == <STLMap {-1000020040: <BDSOutputROOTGeant4Data::ParticleInfo (version 1) at 0x7f6d05752220>, ...} at 0x7f6d057397f0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_40(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue-407.root")) as file:
        branch = file["tree/branch"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0]["fDatime"] == 1749155840
        # py[-1] == <TDatime at 0x7f79c7368430>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_41(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue468.root")) as file:
        branch = file["Event/Trajectory./Trajectory.XYZ"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert len(py[0]) == 0
        # py[-1] == <STLVector [] at 0x7feac87629a0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_42(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue468.root")) as file:
        branch = file["Event/Trajectory./Trajectory.energyDeposit"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert len(py[0]) == 0
        # py[-1] == <STLVector [] at 0x7feac87629a0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_43(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue468.root")) as file:
        branch = file["Event/Trajectory./Trajectory.ionA"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert len(py[0]) == 0
        # py[-1] == <STLVector [] at 0x7f90c6543af0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_44(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue46.root")) as file:
        branch = file["tree/evt"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0]["ArrayBool"][-1] == True
        # py[-1] == <Event (version 1) at 0x7f1e3aef2dc0>
        assert py.layout.form == interp.awkward_form(branch.file)


def test_45():
    with uproot.open(skhep_testdata.data_path("uproot-issue485.root")) as file:
        branch = file["MCTrack/global"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        py = branch.array(interp, library="np", entry_stop=2)
        # py[-1] == <STLVector [<allpix::MCTrack (version 2) at 0x7f88b9a8ad90>, ...] at 0x7f88b9a8ad00>


def test_46():
    with uproot.open(skhep_testdata.data_path("uproot-issue485.root")) as file:
        branch = file["MCParticle/detector1"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        py = branch.array(interp, library="np", entry_stop=2)
        # py[-1] == <STLVector [<allpix::MCParticle (version 6) at 0x7f94bc223760>, ...] at 0x7f94bc223550>


@pytest.mark.parametrize("is_forth", [False, True])
def test_47(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue494.root")) as file:
        branch = file["Geant4Data/Geant4Data./Geant4Data.particles"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[-1][-1]["1"]["mass"] == pytest.approx(3.727379)
        # py[-1] == <STLMap {-1000020040: <BDSOutputROOTGeant4Data::ParticleInfo (version 1) at 0x7f53a43c2220>, ...} at 0x7f53a44278b0>
        assert py.layout.form == interp.awkward_form(branch.file)


def test_48():
    with uproot.open(skhep_testdata.data_path("uproot-issue494.root")) as file:
        branch = file["Geant4Data/Geant4Data./Geant4Data.ions"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        py = branch.array(interp, library="np")
        assert len(py) == 1
        assert len(py[0]) == 0
        # py[-1] == <STLMap {} at 0x7fda831357f0>


def test_49():
    with uproot.open(skhep_testdata.data_path("uproot-issue498.root")) as file:
        branch = file["MCParticle/timepix"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        py = branch.array(interp, library="np", entry_stop=2)
        assert py[0][0].member("particle_id_") == 22
        # py[-1] == <STLVector [<allpix::MCParticle (version 6) at 0x7f0697bf1820>] at 0x7f0697bf1a00>


def test_50():
    with uproot.open(skhep_testdata.data_path("uproot-issue498.root")) as file:
        branch = file["PixelHit/timepix"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        py = branch.array(interp, library="np", entry_stop=2)
        assert len(py[0]) == 0
        assert len(py[1]) == 0
        # py[-1] == <STLVector [] at 0x7f769cbf0bb0>


@pytest.mark.parametrize("is_forth", [False, True])
def test_51(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue510b.root")) as file:
        branch = file["EDepSimEvents/Event/Primaries/Primaries.GeneratorName"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[-1][0] == "GENIE:fixed@density-fixed"
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_52(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue513.root")) as file:
        branch = file["Delphes/EFlowPhoton/EFlowPhoton.Particles"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1][-1]["refs"][-1] == 2223
        # py[-1] == array([<TRefArray [2407] at 0x7f7888dff0d0>, <TRefArray [2411] at 0x7f7888dee100>, <TRefArray [2405] at 0x7f7888dff130>, <TRefArray [2406] at 0x7f7888dff1f0>, <TRefArray [2523] at 0x7f7888dff280>, <TRefArray [2352] at 0x7f7888dff310>, <TRefArray [2192] at 0x7f7888dff3a0>, <TRefArray [2425] at 0x7f7888dff430>, <TRefArray [2340] at 0x7f7888dff4c0>, <TRefArray [2533] at 0x7f7888dff550>, <TRefArray [2264] at 0x7f7888dff5e0>, <TRefArray [2263] at 0x7f7888dff670>, <TRefArray [2396] at 0x7f7888dff700>, <TRefArray [2519] at 0x7f7888dff790>, <TRefArray [2044] at 0x7f7888dff820>, <TRefArray [2273] at 0x7f7888dff8b0>, <TRefArray [2270] at 0x7f7888dff940>, <TRefArray [2388] at 0x7f7888dff9d0>, <TRefArray [2473] at 0x7f7888dffa60>, <TRefArray [2272] at 0x7f7888dffaf0>, <TRefArray [2475] at 0x7f7888dffb80>, <TRefArray [2212] at 0x7f7888dffc10>, <TRefArray [2220] at 0x7f7888dffca0>, <TRefArray [2358] at 0x7f7888dffd30>, <TRefArray [2472] at 0x7f7888dffdc0>, <TRefArray [2359] at 0x7f7888dffe50>, <TRefArray [2360] at 0x7f7888dffee0>, <TRefArray [2201] at 0x7f7888dfff70>, <TRefArray [2362] at 0x7f7888e02040>, <TRefArray [2537] at 0x7f7888e020d0>, <TRefArray [2230] at 0x7f7888e02160>, <TRefArray [2488] at 0x7f7888e021f0>, <TRefArray [2307] at 0x7f7888e02280>, <TRefArray [2570] at 0x7f7888e02310>, <TRefArray [2569] at 0x7f7888e023a0>, <TRefArray [2515] at 0x7f7888e02430>, <TRefArray [2423] at 0x7f7888e024c0>, <TRefArray [2571] at 0x7f7888e02550>, <TRefArray [2578] at 0x7f7888e025e0>, <TRefArray [2386] at 0x7f7888e02670>, <TRefArray [2579] at 0x7f7888e02700>, <TRefArray [2580] at 0x7f7888e02790>, <TRefArray [2556] at 0x7f7888e02820>, <TRefArray [2555] at 0x7f7888e028b0>, <TRefArray [2355] at 0x7f7888e02940>, <TRefArray [2222, 2223] at 0x7f7888e029d0>], dtype=object)
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_53(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue513.root")) as file:
        branch = file["Delphes/Jet/Jet.SoftDroppedSubJet2"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][-1]["fE"] == 0.0
        # py[-1] == array([<TLorentzVector (version 4) at 0x7f6a396bf2b0>, <TLorentzVector (version 4) at 0x7f6a396bf250>, <TLorentzVector (version 4) at 0x7f6a396bf1f0>, <TLorentzVector (version 4) at 0x7f6a396bf190>, <TLorentzVector (version 4) at 0x7f6a396bf130>, <TLorentzVector (version 4) at 0x7f6a396bf0d0>], dtype=object)
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_54(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue513.root")) as file:
        branch = file["Delphes/Jet/Jet.TrimmedP4[5]"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0][-1][4]["fE"] == 0.0
        # py[-1] == array([[<TLorentzVector (version 4) at 0x7f6a02cb3220>, <TLorentzVector (version 4) at 0x7f6a02cb3280>, <TLorentzVector (version 4) at 0x7f6a02cb3fd0>, <TLorentzVector (version 4) at 0x7f6a02cb3f70>, <TLorentzVector (version 4) at 0x7f6a02cb3f10>], [<TLorentzVector (version 4) at 0x7f6a02cb3eb0>, <TLorentzVector (version 4) at 0x7f6a02cb3e50>, <TLorentzVector (version 4) at 0x7f6a02cb3df0>, <TLorentzVector (version 4) at 0x7f6a02cb3d90>, <TLorentzVector (version 4) at 0x7f6a02cb3d30>], [<TLorentzVector (version 4) at 0x7f6a02cb3cd0>, <TLorentzVector (version 4) at 0x7f6a02cb3c70>, <TLorentzVector (version 4) at 0x7f6a02cb3c10>, <TLorentzVector (version 4) at 0x7f6a02cb3bb0>, <TLorentzVector (version 4) at 0x7f6a02cb3b50>], [<TLorentzVector (version 4) at 0x7f6a02cb3af0>, <TLorentzVector (version 4) at 0x7f6a02cb3a90>, <TLorentzVector (version 4) at 0x7f6a02cb3a30>, <TLorentzVector (version 4) at 0x7f6a02cb39d0>, <TLorentzVector (version 4) at 0x7f6a02cb3970>], [<TLorentzVector (version 4) at 0x7f6a02cb3910>, <TLorentzVector (version 4) at 0x7f6a02cb38b0>, <TLorentzVector (version 4) at 0x7f6a02cb3850>, <TLorentzVector (version 4) at 0x7f6a02cb34f0>, <TLorentzVector (version 4) at 0x7f6a02cb3580>], [<TLorentzVector (version 4) at 0x7f6a02cb3430>, <TLorentzVector (version 4) at 0x7f6a02cb33d0>, <TLorentzVector (version 4) at 0x7f6a02cb3370>, <TLorentzVector (version 4) at 0x7f6a02cb32e0>, <TLorentzVector (version 4) at 0x7f6a02cba910>]], dtype=object)
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_55(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-issue519.root")) as file:
        branch = file["testtree/testbranch"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[-1][1][5] == pytest.approx(0.346174418926239)
        # py[-1] == <STLVector [[0.15334472, 0.10603446, 0.004459681, 0.0222604, 0.38413692, 0.39519757], ...] at 0x7f38c4650eb0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_56(is_forth):
    with uproot.open(
        skhep_testdata.data_path("uproot-small-evnt-tree-nosplit.root")
    ) as file:
        branch = file["tree/evt"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1]["SliceU64"][0] == 1
        # py[-1] == <Event (version 1) at 0x7fecdbf61dc0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_57(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/vector_string"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[-1][1] == "two"
        # py[-1] == <STLVector ['one', 'two'] at 0x7fdeeb3f8d90>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_58(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/vector_tstring"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1][1] == "two"
        # py[-1] == <STLVector ['one', 'two'] at 0x7f42edc0c0a0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_59(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/vector_vector_int32"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[-1][1][1] == 2
        # py[-1] == <STLVector [[1], [1, 2]] at 0x7ff36af3c2b0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_60(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/vector_vector_string"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[-1][1][1] == "two"
        # py[-1] == <STLVector [['one'], ['one', 'two']] at 0x7fae23700eb0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_61(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/vector_vector_tstring"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[-1][1][1] == "two"
        # py[-1] == <STLVector [['one'], ['one', 'two']] at 0x7f06ad24c460>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_62(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/vector_set_int32"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[-1][1][1] == 2
        # py[-1] == <STLVector [{1}, {1, 2}] at 0x7f7f97e88880>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_63(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/vector_set_string"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[-1][1][1] == "two"
        # py[-1] == <STLVector [{'one'}, {'one', 'two'}] at 0x7fb29ded7370>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_64(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/set_int32"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[-1][1] == 2
        # py[-1] == <STLSet {1, 2} at 0x7f47a620e310>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_65(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/set_string"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[-1][1] == "two"
        # py[-1] == <STLSet {'one', 'two'} at 0x7f5ffd9ff1c0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_66(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/map_int32_int16"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[-1]["1"][1] == 2
        # py[-1] == <STLMap {1: 1, 2: 2} at 0x7f23be1efbe0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_67(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/map_int32_vector_int16"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1][1]["1"][1] == 2
        # py[-1] == <STLMap {1: [1], 2: [1, 2]} at 0x7f899441d2b0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_68(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/map_int32_vector_string"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1][1]["1"][1] == "two"
        # py[-1] == <STLMap {1: ['one'], 2: ['one', 'two']} at 0x7fd19d3288b0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_69(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/map_int32_set_int16"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1][1]["1"][1] == 2
        # py[-1] == <STLMap {1: {1}, 2: {1, 2}} at 0x7f2b3f1b5fa0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_70(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/map_int32_set_string"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1][1]["1"][1] == "two"
        # py[-1] == <STLMap {1: {'one'}, 2: {'one', 'two'}} at 0x7f4718b237c0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_71(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/map_string_int16"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1][1]["1"] == 2
        # py[-1] == <STLMap {'one': 1, 'two': 2} at 0x7f4179bed1f0>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_72(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/map_string_vector_int16"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1][1]["1"][1] == 2
        # py[-1] == <STLMap {'one': [1], 'two': [1, 2]} at 0x7f26376e9b20>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_73(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/map_string_vector_string"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1][1]["1"][1] == "two"
        # py[-1] == <STLMap {'one': ['one'], 'two': ['one', 'two']} at 0x7f3e45d34910>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_74(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/map_string_set_int16"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1][1]["1"][1] == 2
        # py[-1] == <STLMap {'one': {1}, 'two': {1, 2}} at 0x7f5a94b1c760>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_75(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/map_string_set_string"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1][1]["1"][1] == "two"
        # py[-1] == <STLMap {'one': {'one'}, 'two': {'one', 'two'}} at 0x7f1a1e95aa90>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_76(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/map_int32_vector_vector_int16"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1][1]["1"][1][1] == 2
        # py[-1] == <STLMap {1: [[1]], 2: [[1], [1, 2]]} at 0x7fc98bd7ec10>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_77(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/map_int32_vector_set_int16"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1][1]["1"][1][1] == 2
        # py[-1] == <STLMap {1: [{1}], 2: [{1}, {1, 2}]} at 0x7fcf9b191610>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_78(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/map_string_string"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1, 1, "1"] == "TWO"
        # py[-1] == <STLMap {'one': 'ONE', 'two': 'TWO'} at 0x7f887b27cb20>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_79(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-stl_containers.root")) as file:
        branch = file["tree/map_string_tstring"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[1, 1, "1"] == "TWO"
        # py[-1] == <STLMap {'one': 'ONE', 'two': 'TWO'} at 0x7f4c4e527610>
        assert py.layout.form == interp.awkward_form(branch.file)


@pytest.mark.parametrize("is_forth", [False, True])
def test_80(is_forth):
    with uproot.open(
        skhep_testdata.data_path("uproot-vectorVectorDouble.root")
    ) as file:
        branch = file["t/x"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="np", entry_stop=2)
        assert len(py[0]) == 0
        # py[-1] == <STLVector [[], []] at 0x7f046a6951f0>


@pytest.mark.parametrize("is_forth", [False, True])
def test_81(is_forth):
    with uproot.open(skhep_testdata.data_path("uproot-HZZ-objects.root")) as file:
        branch = file["events/MET"]
        interp = uproot.interpretation.identify.interpretation_of(branch, {}, False)
        interp._forth = is_forth
        py = branch.array(interp, library="ak", entry_stop=2)
        assert py[0]["fY"] == pytest.approx(2.5636332035064697)
        # py[-1] == <STLVector [[], []] at 0x7f046a6951f0>
        assert py.layout.form == interp.awkward_form(branch.file)



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import os

import numpy as np
import pytest

import uproot

ak = pytest.importorskip("awkward")


def test(tmp_path):
    filename = os.path.join(str(tmp_path), "whatever.root")

    with uproot.recreate(filename) as file:
        file["tree"] = {"branch": ak.Array([[1, 2, 3], [4, 5, 6]])}

    with uproot.open(filename) as file:
        assert isinstance(file["tree/branch"].interpretation, uproot.AsJagged)



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import json
import sys
from io import StringIO

import numpy
import pytest
import skhep_testdata

import uproot


def test_streamerless_read():
    with uproot.open(
        skhep_testdata.data_path("uproot-sample-5.23.02-uncompressed.root")
    ) as f:
        assert f["sample"].class_version == 16
        assert isinstance(f["sample"], uproot.TTree)
        assert f["sample"].name == f["sample"].member("fName")
        for x in f["sample"].member("fBranches"):
            assert x.class_version == 11
            assert isinstance(x, uproot.TBranch)
            assert x.name == x.member("fName")
        for x in f["sample"].member("fLeaves"):
            assert x.class_version == 1
        assert f.file._streamers is None

    with uproot.open(
        skhep_testdata.data_path("uproot-sample-5.24.00-uncompressed.root")
    ) as f:
        assert f["sample"].class_version == 16
        assert isinstance(f["sample"], uproot.TTree)
        assert f["sample"].name == f["sample"].member("fName")
        for x in f["sample"].member("fBranches"):
            assert x.class_version == 11
            assert isinstance(x, uproot.TBranch)
            assert x.name == x.member("fName")
        for x in f["sample"].member("fLeaves"):
            assert x.class_version == 1
        assert f.file._streamers is None

    with uproot.open(
        skhep_testdata.data_path("uproot-sample-5.25.02-uncompressed.root")
    ) as f:
        assert f["sample"].class_version == 17
        assert isinstance(f["sample"], uproot.TTree)
        assert f["sample"].name == f["sample"].member("fName")
        for x in f["sample"].member("fBranches"):
            assert x.class_version == 12
            assert isinstance(x, uproot.TBranch)
            assert x.name == x.member("fName")
        for x in f["sample"].member("fLeaves"):
            assert x.class_version == 1
        assert f.file._streamers is None

    with uproot.open(
        skhep_testdata.data_path("uproot-sample-5.26.00-uncompressed.root")
    ) as f:
        assert f["sample"].class_version == 18
        assert isinstance(f["sample"], uproot.TTree)
        assert f["sample"].name == f["sample"].member("fName")
        for x in f["sample"].member("fBranches"):
            assert x.class_version == 12
            assert isinstance(x, uproot.TBranch)
            assert x.name == x.member("fName")
        for x in f["sample"].member("fLeaves"):
            assert x.class_version == 1
        assert f.file._streamers is None

    with uproot.open(
        skhep_testdata.data_path("uproot-sample-5.27.02-uncompressed.root")
    ) as f:
        assert f["sample"].class_version == 18
        assert isinstance(f["sample"], uproot.TTree)
        assert f["sample"].name == f["sample"].member("fName")
        for x in f["sample"].member("fBranches"):
            assert x.class_version == 12
            assert isinstance(x, uproot.TBranch)
            assert x.name == x.member("fName")
        for x in f["sample"].member("fLeaves"):
            assert x.class_version == 1
        assert f.file._streamers is None

    with uproot.open(
        skhep_testdata.data_path("uproot-sample-5.28.00-uncompressed.root")
    ) as f:
        assert f["sample"].class_version == 18
        assert isinstance(f["sample"], uproot.TTree)
        assert f["sample"].name == f["sample"].member("fName")
        for x in f["sample"].member("fBranches"):
            assert x.class_version == 12
            assert isinstance(x, uproot.TBranch)
            assert x.name == x.member("fName")
        for x in f["sample"].member("fLeaves"):
            assert x.class_version == 1
        assert f.file._streamers is None

    with uproot.open(
        skhep_testdata.data_path("uproot-sample-5.29.02-uncompressed.root")
    ) as f:
        assert f["sample"].class_version == 18
        assert isinstance(f["sample"], uproot.TTree)
        assert f["sample"].name == f["sample"].member("fName")
        for x in f["sample"].member("fBranches"):
            assert x.class_version == 12
            assert isinstance(x, uproot.TBranch)
            assert x.name == x.member("fName")
        for x in f["sample"].member("fLeaves"):
            assert x.class_version == 1
        assert f.file._streamers is None

    with uproot.open(
        skhep_testdata.data_path("uproot-sample-5.30.00-uncompressed.root")
    ) as f:
        assert f["sample"].class_version == 19
        assert isinstance(f["sample"], uproot.TTree)
        assert f["sample"].name == f["sample"].member("fName")
        for x in f["sample"].member("fBranches"):
            assert x.class_version == 12
            assert isinstance(x, uproot.TBranch)
            assert x.name == x.member("fName")
        for x in f["sample"].member("fLeaves"):
            assert x.class_version == 1
        assert f.file._streamers is None

    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.08.04-uncompressed.root")
    ) as f:
        assert f["sample"].class_version == 19
        assert isinstance(f["sample"], uproot.TTree)
        assert f["sample"].name == f["sample"].member("fName")
        for x in f["sample"].member("fBranches"):
            assert x.class_version == 12
            assert isinstance(x, uproot.TBranch)
            assert x.name == x.member("fName")
        for x in f["sample"].member("fLeaves"):
            assert x.class_version == 1
        assert f.file._streamers is None

    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.10.05-uncompressed.root")
    ) as f:
        assert f["sample"].class_version == 19
        assert isinstance(f["sample"], uproot.TTree)
        assert f["sample"].name == f["sample"].member("fName")
        for x in f["sample"].member("fBranches"):
            assert x.class_version == 12
            assert isinstance(x, uproot.TBranch)
            assert x.name == x.member("fName")
        for x in f["sample"].member("fLeaves"):
            assert x.class_version == 1
        assert f.file._streamers is None

    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.14.00-uncompressed.root")
    ) as f:
        assert f["sample"].class_version == 20
        assert isinstance(f["sample"], uproot.TTree)
        assert f["sample"].name == f["sample"].member("fName")
        for x in f["sample"].member("fBranches"):
            assert x.class_version == 13
            assert isinstance(x, uproot.TBranch)
            assert x.name == x.member("fName")
        for x in f["sample"].member("fLeaves"):
            assert x.class_version == 1
        assert f.file._streamers is None

    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.16.00-uncompressed.root")
    ) as f:
        assert f["sample"].class_version == 20
        assert isinstance(f["sample"], uproot.TTree)
        assert f["sample"].name == f["sample"].member("fName")
        for x in f["sample"].member("fBranches"):
            assert x.class_version == 13
            assert isinstance(x, uproot.TBranch)
            assert x.name == x.member("fName")
        for x in f["sample"].member("fLeaves"):
            assert x.class_version == 1
        assert f.file._streamers is None

    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.18.00-uncompressed.root")
    ) as f:
        assert f["sample"].class_version == 20
        assert isinstance(f["sample"], uproot.TTree)
        assert f["sample"].name == f["sample"].member("fName")
        for x in f["sample"].member("fBranches"):
            assert x.class_version == 13
            assert isinstance(x, uproot.TBranch)
            assert x.name == x.member("fName")
        for x in f["sample"].member("fLeaves"):
            assert x.class_version == 1
        assert f.file._streamers is None

    with uproot.open(
        skhep_testdata.data_path("uproot-sample-6.20.04-uncompressed.root")
    ) as f:
        assert f["sample"].class_version == 20
        assert isinstance(f["sample"], uproot.TTree)
        assert f["sample"].name == f["sample"].member("fName")
        for x in f["sample"].member("fBranches"):
            assert x.class_version == 13
            assert isinstance(x, uproot.TBranch)
            assert x.name == x.member("fName")
        for x in f["sample"].member("fLeaves"):
            assert x.class_version == 1
        assert f.file._streamers is None

    with uproot.open(skhep_testdata.data_path("uproot-vectorVectorDouble.root")) as f:
        assert f["t"].class_version == 19
        assert f.file._streamers is None


def test_list_streamers():
    with uproot.open(skhep_testdata.data_path("uproot-histograms.root")) as f:
        assert f.file.streamer_dependencies("TNamed") == [
            ("TString", 2),
            ("TObject", 1),
            ("TNamed", 1),
        ]

        output = StringIO()
        f.file.show_streamers("TNamed", stream=output)
        assert (
            output.getvalue()
            == """TString (v2)

TObject (v1)
    fUniqueID: unsigned int (TStreamerBasicType)
    fBits: unsigned int (TStreamerBasicType)

TNamed (v1): TObject (v1)
    fName: TString (TStreamerString)
    fTitle: TString (TStreamerString)
"""
        )



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import os

import numpy as np
import pytest
import skhep_testdata

import uproot
import uproot.writing

ROOT = pytest.importorskip("ROOT")


def test(tmp_path):
    filename = os.path.join(tmp_path, "testy.root")

    f = ROOT.TFile(filename, "recreate")
    x = ROOT.TObjString("hello")
    x.Write()
    f.Close()

    assert uproot.open(filename).classnames() == {"hello;1": "TObjString"}

    with uproot.writing.update(filename) as root_directory:
        root_directory.mkdir("subdir")

    assert uproot.open(filename).classnames() == {
        "hello;1": "TObjString",
        "subdir;1": "TDirectory",
    }

    g = ROOT.TFile(filename, "update")
    g.cd("subdir")
    y = ROOT.TObjString("there")
    y.Write()
    g.Close()

    assert uproot.open(filename).classnames() == {
        "hello;1": "TObjString",
        "subdir;1": "TDirectory",
        "subdir/there;1": "TObjString",
    }



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines a physical layer for remote files, accessed via the XRootD protocol.

Defines a :doc:`uproot.source.xrootd.XRootDResource` (``XRootD.File``) and two
sources: :doc:`uproot.source.xrootd.MultithreadedXRootDSource` and
:doc:`uproot.source.xrootd.XRootDSource`. The latter requires the server to
support vector-read requests; if not, it automatically falls back to
:doc:`uproot.source.xrootd.MultithreadedXRootDSource`.
"""


import sys

import uproot
import uproot.source.chunk
import uproot.source.futures


def get_server_config(file):
    """
    Args:
        file (``XRootD.client.File``): An XRootD file object.

    Query a XRootD server for its configuration.

    Returns a 2-tuple of integers:

    * ``readv_iov_max``: The maximum number of elements that can be
      requested in a single vector read.
    * ``readv_ior_max``: The maximum number of bytes that can be requested
      per *element* in a vector read.
    """
    # Set from some sensible defaults in case this fails
    readv_iov_max = 1024
    readv_ior_max = 2097136

    XRootD_client = uproot.extras.XRootD_client()

    # Check if the file is stored locally on this machine
    last_url = file.get_property("LastURL")
    last_url = XRootD_client.URL(last_url)
    if last_url.protocol == "file" and last_url.hostid == "localhost":
        # The URL will redirect to a local file where XRootD will split the
        # vector reads into multiple sequential reads so just use the default
        return readv_iov_max, readv_ior_max

    # Find where the data is actually stored
    data_server = file.get_property("DataServer")
    if data_server == "":
        raise NotImplementedError

    data_server = XRootD_client.URL(data_server)
    data_server = f"{data_server.protocol}://{data_server.hostid}/"

    # Use a single query call to avoid doubling the latency
    fs = XRootD_client.FileSystem(data_server)
    status, result = fs.query(
        XRootD_client.flags.QueryCode.CONFIG, "readv_iov_max readv_ior_max"
    )
    if status.error:
        raise OSError(status.message)

    # Result is something like b'178956968\n2097136\n'
    readv_iov_max, readv_ior_max = map(int, result.split(b"\n", 1))

    return readv_iov_max, readv_ior_max


class XRootDResource(uproot.source.chunk.Resource):
    """
    Args:
        file_path (str): A URL of the file to open.
        timeout (None or float): An optional timeout in seconds.

    A :doc:`uproot.source.chunk.Resource` for XRootD connections.
    """

    def __init__(self, file_path, timeout):
        self._file_path = file_path
        self._timeout = timeout
        self._open()

    def _open(self):
        XRootD_client = uproot.extras.XRootD_client()

        self._file = XRootD_client.File()

        status, dummy = self._file.open(self._file_path, timeout=self._xrd_timeout())
        if status.error:
            self._xrd_error(status)

    def __getstate__(self):
        state = dict(self.__dict__)
        state.pop("_file")
        return state

    def __setstate__(self, state):
        self.__dict__ = state
        self._open()

    def _xrd_timeout(self):
        if self._timeout is None:
            return 0
        else:
            return int(self._timeout)

    def _xrd_error(self, status):
        self._file.close(timeout=self._xrd_timeout())

        # https://github.com/xrootd/xrootd/blob/8e91462e76ab969720b40fc324714b84e0b4bd42/src/XrdCl/XrdClStatus.hh#L47-L103
        # https://github.com/xrootd/xrootd/blob/250eced4d3787c2ac5be2c8c922134153bbf7f08/src/XrdCl/XrdClStatus.cc#L34-L74
        if status.code in (101, 304, 400):
            raise uproot._util._file_not_found(self._file_path, status.message)

        else:
            raise OSError(
                """XRootD error: {}
in file {}""".format(
                    status.message, self._file_path
                )
            )

    @property
    def timeout(self):
        """
        The timeout in seconds or None.
        """
        return self._timeout

    @property
    def file(self):
        """
        The ``XRootD.client.File`` object.
        """
        return self._file

    @property
    def num_bytes(self):
        status, info = self._file.stat(self._xrd_timeout())
        if status.error:
            self._xrd_error(status)
        return info.size

    @property
    def closed(self):
        return not self._file.is_open()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self._file.close(timeout=self._xrd_timeout())

    def get(self, start, stop):
        """
        Args:
            start (int): Seek position of the first byte to include.
            stop (int): Seek position of the first byte to exclude
                (one greater than the last byte to include).

        Returns a Python buffer of data between ``start`` and ``stop``.
        """
        status, data = self._file.read(
            int(start), int(stop - start), timeout=self._xrd_timeout()
        )
        if status.error:
            self._xrd_error(status)
        return data

    @staticmethod
    def future(source, start, stop):
        """
        Args:
            source (:doc:`uproot.source.xrootd.MultithreadedXRootDSource`): The
                data source.
            start (int): Seek position of the first byte to include.
            stop (int): Seek position of the first byte to exclude
                (one greater than the last byte to include).

        Returns a :doc:`uproot.source.futures.ResourceFuture` that calls
        :ref:`uproot.source.xrootd.XRootDResource.get` with ``start`` and
        ``stop``.
        """

        def task(resource):
            return resource.get(start, stop)

        return uproot.source.futures.ResourceFuture(task)

    @staticmethod
    def partfuture(results, start, stop):
        """
        Returns a :doc:`uproot.source.futures.ResourceFuture` to simply select
        the ``(start, stop)`` item from the ``results`` dict.

        In :ref:`uproot.source.xrootd.XRootDSource.chunks`, each chunk has a
        :ref:`uproot.source.xrootd.XRootDResource.partfuture` that are collectively
        filled by callbacks from :ref:`uproot.source.xrootd.XRootDResource.callbacker`.
        """

        def task(resource):
            return results[start, stop]

        return uproot.source.futures.ResourceFuture(task)

    @staticmethod
    def mergefuture(partfutures):
        """
        Returns a :doc:`uproot.source.futures.ResourceFuture` that merges the
        chunks previously submitted via
        :ref:`uproot.source.xrootd.XRootDResource.partfuture` which had to be split
        """

        def task(resource):
            if len(partfutures) == 1:
                return partfutures[0].result()

            chunk_buffers = []
            for future in partfutures:
                chunk_buffers.append(future.result())
            return b"".join(chunk_buffers)

        return uproot.source.futures.ResourceFuture(task)

    def callbacker(self, futures, results):
        """
        Returns an XRootD callback function to fill the ``futures`` and
        ``results``.
        """

        def callback(status, response, hosts):
            if status.error:
                try:
                    self._xrd_error(status)
                except Exception:
                    excinfo = sys.exc_info()
                    for future in futures.values():
                        future._set_excinfo(excinfo)
            else:
                for chunk in response.chunks:
                    start, stop = chunk.offset, chunk.offset + chunk.length
                    results[start, stop] = chunk.buffer
                    futures[start, stop]._run(None)

        return callback


class XRootDSource(uproot.source.chunk.Source):
    """
    Args:
        file_path (str): A URL of the file to open.
        options: Must include ``"timeout"``, ``"max_num_elements"`` and ``"num_workers"``

    A :doc:`uproot.source.chunk.Source` that uses XRootD's vector-read
    to get many chunks in one request.
    """

    ResourceClass = XRootDResource

    def __init__(self, file_path, **options):
        self._timeout = options["timeout"]
        self._desired_max_num_elements = options["max_num_elements"]
        self._num_workers = options["num_workers"]
        self._num_requests = 0
        self._num_requested_chunks = 0
        self._num_requested_bytes = 0

        self._file_path = file_path
        self._num_bytes = None
        self._open()

    def _open(self):
        self._resource = XRootDResource(self._file_path, self._timeout)

        # this ThreadPool does not need a resource, it's only used to submit
        # futures that wait for chunks that have been split to merge them.
        self._executor = uproot.source.futures.ResourceThreadPoolExecutor(
            [None for i in range(self._num_workers)]
        )

        self._max_num_elements, self._max_element_size = get_server_config(
            self._resource.file
        )
        if self._desired_max_num_elements is not None:
            self._max_num_elements = min(
                self._max_num_elements, self._desired_max_num_elements
            )

    def __getstate__(self):
        state = dict(self.__dict__)
        state.pop("_executor")
        return state

    def __setstate__(self, state):
        self.__dict__ = state
        self._open()

    def __repr__(self):
        path = repr(self._file_path)
        if len(self._file_path) > 10:
            path = repr("..." + self._file_path[-10:])
        return f"<{type(self).__name__} {path} at 0x{id(self):012x}>"

    def chunk(self, start, stop):
        self._num_requests += 1
        self._num_requested_chunks += 1
        self._num_requested_bytes += stop - start

        data = self._resource.get(start, stop)
        future = uproot.source.futures.TrivialFuture(data)
        return uproot.source.chunk.Chunk(self, start, stop, future)

    def chunks(self, ranges, notifications):
        self._num_requests += 1
        self._num_requested_chunks += len(ranges)
        self._num_requested_bytes += sum(stop - start for start, stop in ranges)

        # ranges for xrootd vector reads
        all_request_ranges = [[]]

        # dictionary telling us which xrootd request ranges correspond to the
        # actually requested ranges (given as (start, stop))
        # this is to track which requests were split into smaller ranges and have to be merged
        sub_ranges = {}

        def add_request_range(start, length, sub_ranges_list):
            if len(all_request_ranges[-1]) >= self._max_num_elements:
                all_request_ranges.append([])
            all_request_ranges[-1].append((start, length))
            sub_ranges_list.append((start, start + length))

        # figure out the vector read ranges
        for start, stop in ranges:
            length = stop - start
            sub_ranges[start, stop] = []

            # if range larger than maximum, split into smaller ranges
            if length > self._max_element_size:
                nsplit = length // self._max_element_size
                rem = length % self._max_element_size
                for i in range(nsplit):
                    add_request_range(
                        start + i * self._max_element_size,
                        self._max_element_size,
                        sub_ranges[start, stop],
                    )
                if rem > 0:
                    add_request_range(
                        start + nsplit * self._max_element_size,
                        rem,
                        sub_ranges[start, stop],
                    )
            else:
                add_request_range(start, length, sub_ranges[start, stop])

        # submit the xrootd vector reads
        global_futures = {}
        for _, request_ranges in enumerate(all_request_ranges):
            futures = {}
            results = {}
            for start, size in request_ranges:
                stop = start + size
                partfuture = self.ResourceClass.partfuture(results, start, stop)
                futures[start, stop] = partfuture
                global_futures[start, stop] = partfuture

            callback = self._resource.callbacker(futures, results)

            status = self._resource.file.vector_read(
                chunks=request_ranges,
                timeout=self._resource._xrd_timeout(),
                callback=callback,
            )
            if status.error:
                self._resource._xrd_error(status)

        # create chunks (possibly merging xrootd chunks)
        chunks = []
        for start, stop in ranges:
            partfutures = []
            for sub_start, sub_stop in sub_ranges[start, stop]:
                partfutures.append(global_futures[sub_start, sub_stop])
            future = self.ResourceClass.mergefuture(partfutures)
            chunk = uproot.source.chunk.Chunk(self, start, stop, future)
            future._set_notify(uproot.source.chunk.notifier(chunk, notifications))
            self._executor.submit(future)
            chunks.append(chunk)

        return chunks

    @property
    def resource(self):
        """
        The :doc:`uproot.source.xrootd.XRootDResource` object.
        """
        return self._resource

    @property
    def timeout(self):
        """
        The timeout in seconds or None.
        """
        return self._timeout

    @property
    def file(self):
        """
        The ``XRootD.client.File`` object.
        """
        return self._resource.file

    @property
    def closed(self):
        return self._resource.closed

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        self._executor.shutdown()
        self._resource.__exit__(exception_type, exception_value, traceback)

    @property
    def num_bytes(self):
        if self._num_bytes is None:
            self._num_bytes = self._resource.num_bytes
        return self._num_bytes


class MultithreadedXRootDSource(uproot.source.chunk.MultithreadedSource):
    """
    Args:
        file_path (str): A URL of the file to open.
        options: Must include ``"num_workers"`` and ``"timeout"``.

    A :doc:`uproot.source.chunk.MultithreadedSource` that manages many
    :doc:`uproot.source.xrootd.XRootDResource` objects.
    """

    ResourceClass = XRootDResource

    def __init__(self, file_path, **options):
        self._num_workers = options["num_workers"]
        self._timeout = options["timeout"]
        self._num_requests = 0
        self._num_requested_chunks = 0
        self._num_requested_bytes = 0

        self._file_path = file_path
        self._num_bytes = None
        self._open()

    def _open(self):
        self._executor = uproot.source.futures.ResourceThreadPoolExecutor(
            [
                XRootDResource(self._file_path, self._timeout)
                for x in range(self._num_workers)
            ]
        )

    def __getstate__(self):
        state = dict(self.__dict__)
        state.pop("_executor")
        return state

    def __setstate__(self, state):
        self.__dict__ = state
        self._open()

    @property
    def timeout(self):
        """
        The timeout in seconds or None.
        """
        return self._timeout

    @property
    def num_bytes(self):
        if self._num_bytes is None:
            self._num_bytes = self._executor.workers[0].resource.num_bytes
        return self._num_bytes



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import sys

import numpy
import pytest
import skhep_testdata

import uproot


def test():
    with uproot.open(skhep_testdata.data_path("uproot-nesteddirs.root")) as directory:
        assert directory.keys() == [
            "one;1",
            "one/two;1",
            "one/two/tree;1",
            "one/tree;1",
            "three;1",
            "three/tree;1",
        ]

        assert directory.path == ()
        assert directory["one"].path == ("one",)
        assert directory["one/two"].path == ("one", "two")
        assert directory["three;1"].path == ("three",)

        assert len(directory) == 6
        assert len(directory["one"]) == 3

        assert "one;1/two;1" in directory

        with pytest.raises(KeyError):
            directory["whatever"]

        assert directory.get("whatever") is None



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines the behaviors of ``TGraphErrors``.
"""


class TGraphErrors:
    """
    Behaviors for TGraphErrors: get values and errors as NumPy arrays.
    """

    # TO DO!



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines versioned models for ``TTree``.

See :doc:`uproot.behaviors.TBranch` for definitions of ``TTree``-reading
functions.
"""


import struct
from collections import namedtuple
from enum import Enum

import numpy

import uproot

_tdataset_format1 = struct.Struct(">I")


class Model_TDataSet(uproot.model.Model):
    """
    A :doc:`uproot.model.Model` for ``TDataSet``.
    """

    def read_members(self, chunk, cursor, context, file):
        context["cancel_forth"] = True
        if self.is_memberwise:
            raise NotImplementedError(
                """memberwise serialization of {}
in file {}""".format(
                    type(self).__name__, self.file.file_path
                )
            )

        self._bases.append(
            uproot.models.TNamed.Model_TNamed.read(
                chunk,
                cursor,
                context,
                file,
                self._file,
                self._parent,
                concrete=self.concrete,
            )
        )

        self._members["fParent"] = uproot.deserialization.read_object_any(
            chunk, cursor, context, file, self._file, self.concrete
        )
        self._members["fList"] = uproot.deserialization.read_object_any(
            chunk, cursor, context, file, self._file, self.concrete
        )

    base_names_versions = [
        ("TNamed", 1),
    ]
    member_names = ["fParent", "fList"]
    class_flags = {}
    class_code = None


class EColumnType(Enum):
    """
    An :doc:`Enum` of possible TTable column types.
    """

    kNAN = 0
    kFloat = 1
    kInt = 2
    kLong = 3
    kShort = 4
    kDouble = 5
    kUInt = 6
    kULong = 7
    kUShort = 8
    kUChar = 9
    kChar = 10
    kPtr = 11
    kBool = 12


format = {
    EColumnType.kFloat: struct.Struct(">f"),
    EColumnType.kInt: struct.Struct(">i"),
    EColumnType.kLong: struct.Struct(">q"),
    EColumnType.kShort: struct.Struct(">h"),
    EColumnType.kDouble: struct.Struct(">d"),
    EColumnType.kUInt: struct.Struct(">I"),
    EColumnType.kULong: struct.Struct(">Q"),
    EColumnType.kUShort: struct.Struct(">H"),
    EColumnType.kChar: struct.Struct("c"),
    EColumnType.kBool: struct.Struct("?"),
}

_dtype = {
    EColumnType.kFloat: numpy.dtype(">f4"),
    EColumnType.kInt: numpy.dtype(">i4"),
    EColumnType.kLong: numpy.dtype(">i8"),
    EColumnType.kShort: numpy.dtype(">i2"),
    EColumnType.kDouble: numpy.dtype(">d"),
    EColumnType.kUInt: numpy.dtype(">u4"),
    EColumnType.kULong: numpy.dtype(">u8"),
    EColumnType.kUShort: numpy.dtype(">u2"),
    EColumnType.kChar: numpy.dtype("u1"),
    EColumnType.kBool: numpy.dtype("?"),
}

tableDescriptor_st = namedtuple(
    "tableDescriptor_st",
    [
        "fColumnName",
        "fIndexArray0",
        "fIndexArray1",
        "fIndexArray2",
        "fOffset",
        "fSize",
        "fTypeSize",
        "fDimensions",
        "fType",
    ],
)


_ttabledescriptor4_format1 = struct.Struct(">iqq")


class Model_TTableDescriptor_v4(uproot.model.VersionedModel):
    """
    A :doc:`uproot.model.VersionedModel` for ``TTableDescriptor`` version 4.
    """

    def read_members(self, chunk, cursor, context, file):
        if self.is_memberwise:
            raise NotImplementedError(
                """memberwise serialization of {}
in file {}""".format(
                    type(self).__name__, self.file.file_path
                )
            )

        self._bases.append(
            Model_TDataSet.read(
                chunk,
                cursor,
                context,
                file,
                self._file,
                self._parent,
                concrete=self.concrete,
            )
        )

        (
            self._members["fN"],
            self._members["fMaxIndex"],
            self._members["fSize"],
        ) = cursor.fields(chunk, _ttabledescriptor4_format1, context)

        self._columns = []
        for _ in range(self._members["fMaxIndex"]):
            column = tableDescriptor_st(
                *cursor.fields(chunk, struct.Struct(">32s3iiiiii"), context)
            )
            column = column._replace(
                fColumnName=column.fColumnName.rstrip(b"\x00").decode("utf-8"),
                fType=EColumnType(column.fType),
            )
            self._columns.append(column)

    base_names_versions = []
    member_names = ["fN", "fMaxIndex", "fSize"]
    class_flags = {"has_read_object_any": True}
    class_code = None


class Model_TTableDescriptor(uproot.model.DispatchByVersion):
    """
    A :doc:`uproot.model.DispatchByVersion` for ``TTableDescriptor``.
    """

    known_versions = {4: Model_TTableDescriptor_v4}


_ttable4_format1 = struct.Struct(">iqq")


class Model_TTable_v4(uproot.model.VersionedModel):
    """
    A :doc:`uproot.model.VersionedModel` for ``TTable`` version 4.
    """

    def read_members(self, chunk, cursor, context, file):
        if self.is_memberwise:
            raise NotImplementedError(
                """memberwise serialization of {}
in file {}""".format(
                    type(self).__name__, self.file.file_path
                )
            )

        ioDescriptor = uproot.deserialization.read_object_any(
            chunk, cursor, context, file, self._file, self.concrete
        )

        self._bases.append(
            Model_TDataSet.read(
                chunk,
                cursor,
                context,
                file,
                self._file,
                self._parent,
                concrete=self.concrete,
            )
        )

        (
            self._members["fN"],
            self._members["fMaxIndex"],
            self._members["fSize"],
        ) = cursor.fields(chunk, _ttable4_format1, context)

        assert sum(col.fSize for col in ioDescriptor._columns) == self._members["fSize"]

        buf = cursor.bytes(
            chunk, self._members["fSize"] * self._members["fMaxIndex"], context
        )

        def getFormat(col):
            dtype = _dtype[col.fType]
            if col.fDimensions:
                dtype = (dtype, (col.fSize // _dtype[col.fType].itemsize,))
            return dtype

        dtype = numpy.dtype(
            dict(
                names=[col.fColumnName for col in ioDescriptor._columns],
                formats=[getFormat(col) for col in ioDescriptor._columns],
                offsets=[col.fOffset for col in ioDescriptor._columns],
                itemsize=self._members["fSize"],
            )
        )
        self._data = numpy.frombuffer(buf, dtype=dtype)

    @property
    def data(self):
        view = self._data.view()
        view.flags.writeable = False
        return view

    base_names_versions = []
    member_names = ["fN", "fMaxIndex", "fSize"]
    class_flags = {"has_read_object_any": True}
    class_code = None


class Model_TTable(uproot.model.DispatchByVersion):
    """
    A :doc:`uproot.model.DispatchByVersion` for ``TTable``.
    """

    known_versions = {4: Model_TTable_v4}


uproot.classes["TDataSet"] = Model_TDataSet
uproot.classes["TTableDescriptor"] = Model_TTableDescriptor
uproot.classes["TTable"] = Model_TTable



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module represents external libraries that define "array-like" types so that users can
choose an output format.

The :doc:`uproot.interpretation.library.NumPy` library always works (NumPy is
Uproot's only strict dependency) and outputs NumPy arrays for single arrays
and dict/tuple/list as groups. Objects and jagged arrays are not efficiently
represented, but it provides a zero-dependency least common denominator.

The :doc:`uproot.interpretation.library.Awkward` library is the default and
depends on Awkward Array (``awkward``). It is usually the best option, as it
was designed for Uproot.

The :doc:`uproot.interpretation.library.Pandas` library outputs
``pandas.Series`` for single arrays and ``pandas.DataFrame`` as groups. Objects
are not efficiently represented, but some jagged arrays are encoded as
``pandas.MultiIndex``.

Lazy arrays (:doc:`uproot.behaviors.TBranch.lazy`) can only use the
:doc:`uproot.interpretation.library.Awkward` library.
"""


import gc
import json

import numpy

import uproot


def _rename(name, context):
    if context is None or "rename" not in context:
        return name
    else:
        return context["rename"]


class Library:
    """
    Abstract superclass of array-library handlers, for libraries such as NumPy,
    Awkward Array, and Pandas.

    A library is used in the finalization and grouping stages of producing an
    array, converting it from internal representations like
    :doc:`uproot.interpretation.jagged.JaggedArray`,
    :doc:`uproot.interpretation.strings.StringArray`, and
    :doc:`uproot.interpretation.objects.ObjectArray` into the library's
    equivalents. It can also be required for concatenation and other late-stage
    operations on the output arrays.

    Libraries are usually selected by a string name. These names are held in a
    private registry in the :doc:`uproot.interpretation.library` module.
    """

    @property
    def imported(self):
        """
        Attempts to import the library and returns the imported module.
        """
        raise AssertionError

    def empty(self, shape, dtype):
        """
        Args:
            shape (tuple of int): NumPy array ``shape``. (The first item must
                be zero.)
            dtype (``numpy.dtype`` or its constructor argument): NumPy array
                ``dtype``.

        Returns an empty NumPy-like array.
        """
        return numpy.empty(shape, dtype)

    def zeros(self, shape, dtype):
        """
        Args:
            shape (tuple of int): NumPy array ``shape``. (The first item must
                be zero.)
            dtype (``numpy.dtype`` or its constructor argument): NumPy array
                ``dtype``.

        Returns a NumPy-like array of zeros.
        """
        return numpy.zeros(shape, dtype)

    def finalize(
        self, array, branch, interpretation, entry_start, entry_stop, interp_options
    ):
        """
        Args:
            array (array): Internal, temporary, trimmed array. If this is a
                NumPy array, it may be identical to the output array.
            branch (:doc:`uproot.behaviors.TBranch.TBranch`): The ``TBranch``
                that is represented by this array.
            interpretation (:doc:`uproot.interpretation.Interpretation`): The
                interpretation that produced the ``array``.
            entry_start (int): First entry that is included in the output.
            entry_stop (int): FIrst entry that is excluded (one greater than
                the last entry that is included) in the output.
            interp_options (dict): Flags and other options passed through the
                interpretation process.

        Create a library-appropriate output array for this temporary ``array``.

        This array would represent one ``TBranch`` (i.e. not a "group").
        """
        raise AssertionError

    def group(self, arrays, expression_context, how):
        """
        Args:
            arrays (dict of str \u2192 array): Mapping from names to finalized
                array objets to combine into a group.
            expression_context (list of (str, dict) tuples): Expression strings
                and a dict of metadata about each.
            how (None, str, or container type): Library-dependent instructions
                for grouping. The only recognized container types are ``tuple``,
                ``list``, and ``dict``. Note that the container *type itself*
                must be passed as ``how``, not an instance of that type (i.e.
                ``how=tuple``, not ``how=()``).

        Combine the finalized ``arrays`` into a library-appropriate group type.
        """
        if how is tuple:
            return tuple(arrays[name] for name, _ in expression_context)
        elif how is list:
            return [arrays[name] for name, _ in expression_context]
        elif how is dict or how is None:
            return {_rename(name, c): arrays[name] for name, c in expression_context}
        else:
            raise TypeError(
                "for library {}, how must be tuple, list, dict, or None (for "
                "dict)".format(self.name)
            )

    def global_index(self, array, global_offset):
        """
        Args:
            array (array): The library-appropriate array whose global index
                needs adjustment.
            global_offset (int): A number to add to the global index of
                ``array`` to correct it.

        Apply *in-place* corrections to the global index of ``array`` by adding
        ``global_offset``.

        Even though the operation is performed *in-place*, this method returns
        the ``array``.
        """
        return array

    def concatenate(self, all_arrays):
        """
        Args:
            all_arrays (list of arrays): A list of library-appropriate arrays
                that need to be concatenated.

        Returns a concatenated version of ``all_arrays``.
        """
        raise AssertionError

    def __repr__(self):
        return repr(self.name)

    def __eq__(self, other):
        return type(_libraries[self.name]) is type(_libraries[other.name])  # noqa: E721


class NumPy(Library):
    """
    A :doc:`uproot.interpretation.library.Library` that presents ``TBranch``
    data as NumPy arrays. The standard name for this library is ``"np"``.

    The single-``TBranch`` form for this library is a ``numpy.ndarray``. If
    the data are non-numerical, they will be converted into Python objects and
    stored in an array with ``dtype="O"``. This is inefficient, but it is the
    minimal-dependency option for Python.

    The "group" behavior for this library is:

    * ``how=dict`` or ``how=None``: a dict of str \u2192 array, mapping the
      names to arrays.
    * ``how=tuple``: a tuple of arrays, in the order requested. (Names are
      lost.)
    * ``how=list``: a list of arrays, in the order requested. (Names are lost.)

    Since NumPy arrays are not indexed, ``global_index`` has no effect.
    """

    name = "np"

    @property
    def imported(self):
        import numpy

        return numpy

    def finalize(self, array, branch, interpretation, entry_start, entry_stop, options):
        if isinstance(array, uproot.interpretation.jagged.JaggedArray) and isinstance(
            array.content,
            uproot.interpretation.objects.StridedObjectArray,
        ):
            out = numpy.zeros(len(array), dtype=object)
            for i, x in enumerate(array):
                out[i] = numpy.zeros(x.shape, dtype=object)
                for j, y in x.ndenumerate():
                    out[i][j] = y
            return out

        elif isinstance(
            array,
            uproot.interpretation.objects.StridedObjectArray,
        ):
            out = numpy.zeros(array.shape, dtype=object)
            for i, x in array.ndenumerate():
                out[i] = x
            return out

        elif isinstance(
            array,
            (
                uproot.interpretation.jagged.JaggedArray,
                uproot.interpretation.strings.StringArray,
                uproot.interpretation.objects.ObjectArray,
            ),
        ):
            out = numpy.zeros(len(array), dtype=object)
            for i, x in enumerate(array):
                out[i] = x
            return out

        else:
            return array

    def concatenate(self, all_arrays):
        if len(all_arrays) == 0:
            return all_arrays

        if isinstance(all_arrays[0], (tuple, list)):
            keys = range(len(all_arrays[0]))
        elif isinstance(all_arrays[0], dict):
            keys = list(all_arrays[0])
        else:
            raise AssertionError(repr(all_arrays[0]))

        to_concatenate = {k: [] for k in keys}
        for arrays in all_arrays:
            for k in keys:
                to_concatenate[k].append(arrays[k])

        concatenated = {k: numpy.concatenate(to_concatenate[k]) for k in keys}

        if isinstance(all_arrays[0], tuple):
            return tuple(concatenated[k] for k in keys)
        elif isinstance(all_arrays[0], list):
            return [concatenated[k] for k in keys]
        elif isinstance(all_arrays[0], dict):
            return concatenated


def _strided_to_awkward(awkward, path, interpretation, data):
    contents = []
    names = []
    data = data.flatten()
    for name, member in interpretation.members:
        if not name.startswith("@"):
            p = name
            if len(path) != 0:
                p = path + "/" + name
            if isinstance(member, uproot.interpretation.objects.AsStridedObjects):
                contents.append(_strided_to_awkward(awkward, p, member, data))
            else:
                contents.append(
                    awkward.from_numpy(
                        numpy.array(data[p]), regulararray=True, highlevel=False
                    )
                )
            names.append(name)
    parameters = {
        "__record__": uproot.model.classname_decode(interpretation.model.__name__)[0]
    }
    length = len(data) if len(contents) == 0 else None
    out = awkward.contents.RecordArray(contents, names, length, parameters=parameters)
    for dim in reversed(interpretation.inner_shape):
        out = awkward.contents.RegularArray(out, dim)
    return out


def _object_to_awkward_json(form, obj):
    if form["class"] == "NumpyArray":
        return obj

    elif form["class"] == "RecordArray":
        out = {}
        for name, subform in zip(form["fields"], form["contents"]):
            if not name.startswith("@"):
                if obj.has_member(name):
                    out[name] = _object_to_awkward_json(subform, obj.member(name))
                else:
                    out[name] = _object_to_awkward_json(subform, getattr(obj, name))
        return out

    elif form["class"][:15] == "ListOffsetArray":
        if form["parameters"].get("__array__") == "string":
            return obj

        elif form["content"]["parameters"].get("__array__") == "sorted_map":
            key_form = form["content"]["contents"][0]
            value_form = form["content"]["contents"][1]
            return [
                (
                    _object_to_awkward_json(key_form, x),
                    _object_to_awkward_json(value_form, y),
                )
                for x, y in obj.items()
            ]

        else:
            subform = form["content"]
            return [_object_to_awkward_json(subform, x) for x in obj]

    elif form["class"] == "RegularArray":
        subform = form["content"]
        return [_object_to_awkward_json(subform, x) for x in obj]

    else:
        raise AssertionError(form["class"])


def _awkward_offsets(awkward, form, array):
    if isinstance(array, awkward.contents.EmptyArray):
        if form["offsets"] == "i32":
            return awkward.index.Index32(numpy.zeros(1, dtype=numpy.int32))
        elif form["offsets"] == "u32":
            return awkward.index.IndexU32(numpy.zeros(1, dtype=numpy.uint32))
        elif form["offsets"] == "i64":
            return awkward.index.Index64(numpy.zeros(1, dtype=numpy.int64))
        else:
            raise AssertionError(form["offsets"])
    else:
        if form["offsets"] == "i32":
            return awkward.index.Index32(
                numpy.asarray(array.offsets, dtype=numpy.int32)
            )
        elif form["offsets"] == "u32":
            return awkward.index.IndexU32(
                numpy.asarray(array.offsets, dtype=numpy.uint32)
            )
        elif form["offsets"] == "i64":
            return awkward.index.Index64(
                numpy.asarray(array.offsets, dtype=numpy.int64)
            )
        else:
            raise AssertionError(form["offsets"])


def _awkward_json_to_array(awkward, form, array):
    if form["class"] == "NumpyArray":
        form = awkward.forms.from_json(json.dumps(form))
        dtype = awkward.types.numpytype.primitive_to_dtype(form.primitive)
        if isinstance(array, awkward.contents.EmptyArray):
            return awkward.contents.NumpyArray(
                numpy.empty(0, dtype=dtype),
                parameters=form.parameters,
            )
        else:
            return awkward.contents.NumpyArray(
                numpy.asarray(array.data, dtype=dtype),
                parameters=form.parameters,
            )

    elif form["class"] == "RecordArray":
        contents = []
        names = []
        for name, subform in zip(form["fields"], form["contents"]):
            if not name.startswith("@"):
                if isinstance(array, awkward.contents.EmptyArray):
                    contents.append(_awkward_json_to_array(awkward, subform, array))
                else:
                    contents.append(
                        _awkward_json_to_array(awkward, subform, array[name])
                    )
                names.append(name)
        length = len(array) if len(contents) == 0 else None
        return awkward.contents.RecordArray(
            contents, names, length, parameters=form["parameters"]
        )

    elif form["class"][:15] == "ListOffsetArray":
        if form["parameters"].get("__array__") == "string":
            if isinstance(array, awkward.contents.EmptyArray):
                content = awkward.contents.NumpyArray(
                    numpy.empty(0, dtype=numpy.uint8),
                    parameters=form["content"]["parameters"],
                )
                return awkward.contents.ListOffsetArray(
                    awkward.index.Index64(numpy.array([0], dtype=numpy.uint8)),
                    content,
                    parameters=form["parameters"],
                )
            else:
                content = _awkward_json_to_array(
                    awkward, form["content"], array.content
                )
                return type(array)(
                    array.offsets, content, parameters=form["parameters"]
                )

        elif form["content"]["parameters"].get("__array__") == "sorted_map":
            offsets = _awkward_offsets(awkward, form, array)
            key_form = form["content"]["contents"][0]
            value_form = form["content"]["contents"][1]
            if isinstance(array, awkward.contents.EmptyArray):
                keys = _awkward_json_to_array(awkward, key_form, array)
                values = _awkward_json_to_array(awkward, value_form, array)
                content = awkward.contents.RecordArray(
                    (keys, values),
                    None,
                    0,
                    parameters=form["content"]["parameters"],
                )
            else:
                keys = _awkward_json_to_array(awkward, key_form, array.content["0"])
                values = _awkward_json_to_array(awkward, value_form, array.content["1"])
                length = len(array.content) if len(keys) == 0 else None
                content = awkward.contents.RecordArray(
                    (keys, values),
                    None,
                    length,
                    parameters=form["content"]["parameters"],
                )
            cls = uproot._util._content_cls_from_name(awkward, form["class"])
            return cls(offsets, content, parameters=form["parameters"])

        else:
            offsets = _awkward_offsets(awkward, form, array)
            if isinstance(array, awkward.contents.EmptyArray):
                content = _awkward_json_to_array(awkward, form["content"], array)
            else:
                content = _awkward_json_to_array(
                    awkward, form["content"], array.content
                )
            cls = uproot._util._content_cls_from_name(awkward, form["class"])
            return cls(offsets, content, parameters=form["parameters"])

    elif form["class"] == "RegularArray":
        if isinstance(array, awkward.contents.EmptyArray):
            content = _awkward_json_to_array(awkward, form["content"], array)
        else:
            content = _awkward_json_to_array(awkward, form["content"], array.content)
        return awkward.contents.RegularArray(
            content, form["size"], parameters=form["parameters"]
        )

    else:
        raise AssertionError(form["class"])


def _awkward_add_doc(awkward, array, branch, ak_add_doc):
    if ak_add_doc:
        return awkward.with_parameter(array, "__doc__", branch.title)
    else:
        return array


class Awkward(Library):
    """
    A :doc:`uproot.interpretation.library.Library` that presents ``TBranch``
    data as Awkward Arrays. The standard name for this library is ``"ak"``.

    This is the default for all functions that require a
    :doc:`uproot.interpretation.library.Library`, though Uproot does not
    explicitly depend on Awkward Array. If you are confronted with a message
    that Awkward Array is not installed, either install ``awkward`` or
    select another library (likely :doc:`uproot.interpretation.library.NumPy`).

    Both the single-``TBranch`` and "group" forms for this library are
    ``ak.Array``, though groups are always arrays of records. Awkward Array
    was originally developed for Uproot, so the data structures are usually
    optimial for Uproot data.

    The "group" behavior for this library is:

    * ``how=None``: an array of Awkward records.
    * ``how=dict``: a dict of str \u2192 array, mapping the names to arrays.
    * ``how=tuple``: a tuple of arrays, in the order requested. (Names are
      lost.)
    * ``how=list``: a list of arrays, in the order requested. (Names are lost.)

    Since Awkward arrays are not indexed, ``global_index`` has no effect.
    """

    name = "ak"

    @property
    def imported(self):
        return uproot.extras.awkward()

    def finalize(self, array, branch, interpretation, entry_start, entry_stop, options):
        awkward = self.imported

        ak_add_doc = options.get("ak_add_doc", False)

        if isinstance(array, awkward.contents.Content):
            return _awkward_add_doc(awkward, awkward.Array(array), branch, ak_add_doc)

        elif isinstance(array, uproot.interpretation.objects.StridedObjectArray):
            return _awkward_add_doc(
                awkward,
                awkward.Array(
                    _strided_to_awkward(awkward, "", array.interpretation, array.array)
                ),
                branch,
                ak_add_doc,
            )

        elif isinstance(array, uproot.interpretation.jagged.JaggedArray) and isinstance(
            array.content, uproot.interpretation.objects.StridedObjectArray
        ):
            content = _strided_to_awkward(
                awkward, "", array.content.interpretation, array.content.array
            )
            if issubclass(array.offsets.dtype.type, numpy.int32):
                offsets = awkward.index.Index32(array.offsets)
                layout = awkward.contents.ListOffsetArray32(offsets, content)
            else:
                offsets = awkward.index.Index64(array.offsets)
                layout = awkward.contents.ListOffsetArray(offsets, content)
            return _awkward_add_doc(awkward, awkward.Array(layout), branch, ak_add_doc)

        elif isinstance(array, uproot.interpretation.jagged.JaggedArray):
            content = awkward.from_numpy(
                array.content, regulararray=True, highlevel=False
            )
            if issubclass(array.offsets.dtype.type, numpy.int32):
                offsets = awkward.index.Index32(array.offsets)
                layout = awkward.contents.ListOffsetArray32(offsets, content)
            else:
                offsets = awkward.index.Index64(array.offsets)
                layout = awkward.contents.ListOffsetArray(offsets, content)
            return _awkward_add_doc(awkward, awkward.Array(layout), branch, ak_add_doc)

        elif isinstance(array, uproot.interpretation.strings.StringArray):
            content = awkward.contents.NumpyArray(
                numpy.frombuffer(array.content, dtype=numpy.dtype(numpy.uint8)),
                parameters={"__array__": "char"},
            )
            if issubclass(array.offsets.dtype.type, numpy.int32):
                offsets = awkward.index.Index32(array.offsets)
                layout = awkward.contents.ListOffsetArray32(
                    offsets, content, parameters={"__array__": "string"}
                )
            elif issubclass(array.offsets.dtype.type, numpy.uint32):
                offsets = awkward.index.IndexU32(array.offsets)
                layout = awkward.contents.ListOffsetArrayU32(
                    offsets, content, parameters={"__array__": "string"}
                )
            elif issubclass(array.offsets.dtype.type, numpy.int64):
                offsets = awkward.index.Index64(array.offsets)
                layout = awkward.contents.ListOffsetArray(
                    offsets, content, parameters={"__array__": "string"}
                )
            else:
                raise AssertionError(repr(array.offsets.dtype))
            return _awkward_add_doc(awkward, awkward.Array(layout), branch, ak_add_doc)

        elif isinstance(interpretation, uproot.interpretation.objects.AsObjects):
            try:
                form = json.loads(
                    interpretation.awkward_form(interpretation.branch.file).to_json()
                )
            except uproot.interpretation.objects.CannotBeAwkward as err:
                raise ValueError(
                    """cannot produce Awkward Arrays for interpretation {} because

    {}

instead, try library="np" instead of library="ak" or globally set uproot.default_library

in file {}
in object {}""".format(
                        repr(interpretation),
                        err.because,
                        interpretation.branch.file.file_path,
                        interpretation.branch.object_path,
                    )
                ) from err

            unlabeled = awkward.from_iter(
                (_object_to_awkward_json(form, x) for x in array), highlevel=False
            )
            return _awkward_add_doc(
                awkward,
                awkward.Array(_awkward_json_to_array(awkward, form, unlabeled)),
                branch,
                ak_add_doc,
            )

        elif array.dtype.names is not None:
            length, shape = array.shape[0], array.shape[1:]
            array = array.reshape(-1)
            contents = []
            for name in array.dtype.names:
                contents.append(
                    awkward.from_numpy(
                        numpy.array(array[name]), regulararray=True, highlevel=False
                    )
                )
            if len(contents) != 0:
                length = None
            out = awkward.contents.RecordArray(contents, array.dtype.names, length)
            for size in shape[::-1]:
                out = awkward.contents.RegularArray(out, size)
            return _awkward_add_doc(awkward, awkward.Array(out), branch, ak_add_doc)

        else:
            return _awkward_add_doc(
                awkward,
                awkward.from_numpy(array, regulararray=True),
                branch,
                ak_add_doc,
            )

    def group(self, arrays, expression_context, how):
        awkward = self.imported

        if how is tuple:
            return tuple(arrays[name] for name, _ in expression_context)
        elif how is list:
            return [arrays[name] for name, _ in expression_context]
        elif how is dict:
            return {_rename(name, c): arrays[name] for name, c in expression_context}
        elif how is None:
            if len(expression_context) == 0:
                return awkward.Array(
                    awkward.contents.RecordArray([], fields=[], length=0)
                )
            else:
                return awkward.Array(
                    {_rename(name, c): arrays[name] for name, c in expression_context}
                )
        elif how == "zip":
            nonjagged = []
            offsets = []
            jaggeds = []
            renamed_arrays = {}
            for name, context in expression_context:
                array = renamed_arrays[_rename(name, context)] = arrays[name]
                if context["is_jagged"]:
                    if (
                        isinstance(array.layout, awkward.contents.ListArray)
                        or array.layout.offsets[0] != 0
                    ):
                        array_layout = array.layout.to_ListOffsetArray64(True)
                    else:
                        array_layout = array.layout
                    if len(offsets) == 0:
                        offsets.append(array_layout.offsets)
                        jaggeds.append([_rename(name, context)])
                    else:
                        for o, j in zip(offsets, jaggeds):
                            if numpy.array_equal(array_layout.offsets, o):
                                j.append(_rename(name, context))
                                break
                        else:
                            offsets.append(array_layout.offsets)
                            jaggeds.append([_rename(name, context)])
                else:
                    nonjagged.append(_rename(name, context))

            out = None
            if len(nonjagged) != 0:
                if len(nonjagged) == 0:
                    out = awkward.Array(
                        awkward.contents.RecordArray([], fields=[], length=0)
                    )
                else:
                    out = awkward.Array(
                        {name: renamed_arrays[name] for name in nonjagged},
                    )
            for number, jagged in enumerate(jaggeds):
                cut = len(jagged[0])
                for name in jagged:
                    cut = min(cut, len(name))
                    while cut > 0 and (
                        name[:cut] != jagged[0][:cut]
                        or name[cut - 1] not in ("_", ".", "/")
                    ):
                        cut -= 1
                    if cut == 0:
                        break
                if (
                    out is not None
                    and cut != 0
                    and jagged[0][:cut].strip("_./") in awkward.fields(out)
                ):
                    cut = 0
                if cut == 0:
                    common = f"jagged{number}"
                    if len(jagged) == 0:
                        subarray = awkward.Array(
                            awkward.contents.RecordArray([], fields=[], length=0)
                        )
                    else:
                        subarray = awkward.zip(
                            {name: renamed_arrays[name] for name in jagged}
                        )
                else:
                    common = jagged[0][:cut].strip("_./")
                    if len(jagged) == 0:
                        subarray = awkward.Array(
                            awkward.contents.RecordArray([], fields=[], length=0)
                        )
                    else:
                        subarray = awkward.zip(
                            {
                                name[cut:].strip("_./"): renamed_arrays[name]
                                for name in jagged
                            }
                        )
                if out is None:
                    out = awkward.Array({common: subarray})
                else:
                    out = awkward.with_field(out, subarray, common)

            return out
        else:
            raise TypeError(
                'for library {}, how must be tuple, list, dict, "zip" for '
                "a record array with jagged arrays zipped, if possible, or "
                "None, for an unzipped record array".format(self.name)
            )

    def concatenate(self, all_arrays):
        awkward = self.imported

        if len(all_arrays) == 0:
            return all_arrays

        if isinstance(all_arrays[0], (tuple, list)):
            keys = range(len(all_arrays[0]))
        elif isinstance(all_arrays[0], dict):
            keys = list(all_arrays[0])
        else:
            return awkward.concatenate(all_arrays)

        to_concatenate = {k: [] for k in keys}
        for arrays in all_arrays:
            for k in keys:
                to_concatenate[k].append(arrays[k])

        concatenated = {k: awkward.concatenate(to_concatenate[k]) for k in keys}

        if isinstance(all_arrays[0], tuple):
            return tuple(concatenated[k] for k in keys)
        elif isinstance(all_arrays[0], list):
            return [concatenated[k] for k in keys]
        elif isinstance(all_arrays[0], dict):
            return concatenated


def _is_pandas_rangeindex(pandas, index):
    if hasattr(pandas, "RangeIndex") and isinstance(index, pandas.RangeIndex):
        return True
    if hasattr(index, "is_integer") and index.is_integer():
        return True
    if uproot._util.parse_version(pandas.__version__) < uproot._util.parse_version(
        "1.4.0"
    ) and isinstance(index, pandas.Int64Index):
        return True

    return False


def _strided_to_pandas(path, interpretation, data, arrays, columns):
    for name, member in interpretation.members:
        if not name.startswith("@"):
            p = path + (name,)
            if isinstance(member, uproot.interpretation.objects.AsStridedObjects):
                _strided_to_pandas(p, member, data, arrays, columns)
            else:
                arrays.append(data["/".join(p)])
                columns.append(p)


def _pandas_basic_index(pandas, entry_start, entry_stop):
    if hasattr(pandas, "RangeIndex"):
        return pandas.RangeIndex(entry_start, entry_stop)
    else:
        return pandas.Int64Index(range(entry_start, entry_stop))


def _pandas_only_series(pandas, original_arrays, expression_context):
    arrays = {}
    names = []
    for name, context in expression_context:
        arrays[_rename(name, context)] = original_arrays[name]
        names.append(_rename(name, context))
    return arrays, names


def _pandas_memory_efficient(pandas, series, names):
    # Pandas copies the data, so at least feed columns one by one
    gc.collect()
    out = None
    for name in names:
        if out is None:
            out = series[name].to_frame(name=name)
        else:
            out[name] = series[name]
        del series[name]
    if out is None:
        return pandas.DataFrame(data=series, columns=names)
    else:
        return out


class Pandas(Library):
    """
    A :doc:`uproot.interpretation.library.Library` that presents ``TBranch``
    data as Pandas Series and DataFrames. The standard name for this library is
    ``"pd"``.

    The single-``TBranch`` (with a single ``TLeaf``) form for this library is
    ``pandas.Series``, and the "group" form is ``pandas.DataFrame``.

    The "group" behavior for this library is:

    * ``how=None`` or a string: passed to ``pandas.merge`` as its ``how``
      parameter, which would be relevant if jagged arrays with different
      multiplicity are requested.
    * ``how=dict``: a dict of str \u2192 array, mapping the names to
      ``pandas.Series``.
    * ``how=tuple``: a tuple of ``pandas.Series``, in the order requested.
      (Names are assigned to the ``pandas.Series``.)
    * ``how=list``: a list of ``pandas.Series``, in the order requested.
      (Names are assigned to the ``pandas.Series``.)

    Pandas Series and DataFrames are indexed, so ``global_index`` adjusts them.
    """

    name = "pd"

    @property
    def imported(self):
        return uproot.extras.pandas()

    def finalize(self, array, branch, interpretation, entry_start, entry_stop, options):
        pandas = self.imported
        index = _pandas_basic_index(pandas, entry_start, entry_stop)

        if (
            isinstance(array, numpy.ndarray)
            and array.dtype.names is None
            and len(array.shape) == 1
        ):
            return pandas.Series(array, index=index)
        else:
            awkward_pandas = uproot.extras.awkward_pandas()
            ak_lib = _libraries[Awkward.name]
            ak_arr = ak_lib.finalize(
                array, branch, interpretation, entry_start, entry_stop, options
            )
            return pandas.Series(
                awkward_pandas.AwkwardExtensionArray(ak_arr), index=index
            )

    def group(self, arrays, expression_context, how):
        pandas = self.imported

        if how is tuple:
            return tuple(arrays[name] for name, _ in expression_context)

        elif how is list:
            return [arrays[name] for name, _ in expression_context]

        elif how is dict:
            return {_rename(name, c): arrays[name] for name, c in expression_context}

        elif uproot._util.isstr(how) or how is None:
            arrays, names = _pandas_only_series(pandas, arrays, expression_context)
            return _pandas_memory_efficient(pandas, arrays, names)

        else:
            raise TypeError(
                "for library {}, how must be tuple, list, dict, str (for "
                "pandas.merge's 'how' parameter, or None (for one or more"
                "DataFrames without merging)".format(self.name)
            )

    def global_index(self, arrays, global_offset):
        if isinstance(arrays, tuple):
            return tuple(self.global_index(x, global_offset) for x in arrays)
        elif isinstance(arrays, list):
            return list(self.global_index(x, global_offset) for x in arrays)

        if type(arrays.index).__name__ == "RangeIndex":
            index_start = arrays.index.start
            index_stop = arrays.index.stop
            arrays.index = type(arrays.index)(
                index_start + global_offset, index_stop + global_offset
            )

        else:
            index = arrays.index.arrays
            numpy.add(index, global_offset, out=index)

        return arrays

    def concatenate(self, all_arrays):
        pandas = self.imported

        if len(all_arrays) == 0:
            return all_arrays

        if isinstance(all_arrays[0], (tuple, list)):
            keys = range(len(all_arrays[0]))
        elif isinstance(all_arrays[0], dict):
            keys = list(all_arrays[0])
        else:
            return pandas.concat(all_arrays)

        to_concatenate = {k: [] for k in keys}
        for arrays in all_arrays:
            for k in keys:
                to_concatenate[k].append(arrays[k])

        concatenated = {k: pandas.concat(to_concatenate[k]) for k in keys}

        if isinstance(all_arrays[0], tuple):
            return tuple(concatenated[k] for k in keys)
        elif isinstance(all_arrays[0], list):
            return [concatenated[k] for k in keys]
        elif isinstance(all_arrays[0], dict):
            return concatenated


_libraries = {
    NumPy.name: NumPy(),
    Awkward.name: Awkward(),
    Pandas.name: Pandas(),
}

_libraries["numpy"] = _libraries[NumPy.name]
_libraries["Numpy"] = _libraries[NumPy.name]
_libraries["NumPy"] = _libraries[NumPy.name]
_libraries["NUMPY"] = _libraries[NumPy.name]

_libraries["awkward1"] = _libraries[Awkward.name]
_libraries["Awkward1"] = _libraries[Awkward.name]
_libraries["AWKWARD1"] = _libraries[Awkward.name]
_libraries["awkward"] = _libraries[Awkward.name]
_libraries["Awkward"] = _libraries[Awkward.name]
_libraries["AWKWARD"] = _libraries[Awkward.name]

_libraries["pandas"] = _libraries[Pandas.name]
_libraries["Pandas"] = _libraries[Pandas.name]
_libraries["PANDAS"] = _libraries[Pandas.name]


def _regularize_library(library):
    if isinstance(library, Library):
        if library.name in _libraries:
            return _libraries[library.name]
        else:
            raise ValueError(
                "library {} ({}) cannot be used in this function".format(
                    type(library).__name__, repr(library.name)
                )
            )

    elif isinstance(library, type) and issubclass(library, Library):
        if library().name in _libraries:
            return _libraries[library().name]
        else:
            raise ValueError(
                "library {} ({}) cannot be used in this function".format(
                    library.__name__, repr(library().name)
                )
            )

    else:
        try:
            return _libraries[library]
        except KeyError as err:
            raise ValueError(
                """library {} not recognized (for this function); """
                """try "np" (NumPy), "ak" (Awkward Array), or "pd" (Pandas) """
                """instead""".format(repr(library))
            ) from err



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import pytest
import skhep_testdata

import uproot


def test():
    with uproot.open(skhep_testdata.data_path("uproot-issue-172.root")) as f:
        t = f["events"]
        t.show()
        assert (
            t["rec_part_px_VecOps"].typename == "std::vector<float>"
        )  # without the allocator
        t["rec_part_px_VecOps"].array(library="np")



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

"""
This module defines the behaviors of ``RooHist``.
"""


import numpy

import uproot
import uproot.behaviors.TGraphAsymmErrors

# ['@fUniqueID', '@fBits', 'fName', 'fTitle', 'fLineColor', 'fLineStyle',
#  'fLineWidth', 'fFillColor', 'fFillStyle', 'fMarkerColor', 'fMarkerStyle',
#  'fMarkerSize', 'fNpoints', 'fX', 'fY', 'fFunctions', 'fHistogram',
#  'fMinimum', 'fMaximum', 'fEXlow', 'fEXhigh', 'fEYlow', 'fEYhigh']


class RooHist(uproot.behaviors.TGraphAsymmErrors.TGraphAsymmErrors):
    """
    Behavior for ``RooHist``.

    This consists of ``to_boost`` (and ``to_hist``) and ``to_numpy``
    providing access to the histogram.
    """

    @property
    def name(self):
        """
        The name of the histogram.
        """
        return self.member("fName")

    @property
    def title(self):
        """
        The title of the histogram.
        """
        return self.member("fTitle")

    def to_numpy(self, dd=False):
        """
        Args:
            dd (bool): If True, the return type follows
                `numpy.histogramdd <https://numpy.org/doc/stable/reference/generated/numpy.histogramdd.html>`__;
                otherwise, it follows `numpy.histogram <https://numpy.org/doc/stable/reference/generated/numpy.histogram.html>`__
                and `numpy.histogram2d <https://numpy.org/doc/stable/reference/generated/numpy.histogram2d.html>`__.

        Converts the ``RooHist`` into a form like the ones produced by the NumPy
        histogram functions.
        """
        bin_centers, values = self.values()
        bin_edges_low = bin_centers - self.errors(which="low", axis="x")
        bin_edges_high = bin_centers + self.errors(which="high", axis="x")
        if not numpy.all(numpy.isclose(bin_edges_low[1:], bin_edges_high[:-1])):
            raise ValueError("bin_edges_low[1:] != bin_edges_high[:-1]")
        bin_edges = numpy.append(bin_edges_low, [bin_edges_high[-1]])
        if dd:
            return values, (bin_edges,)
        else:
            return values, bin_edges

    def to_boost(self):
        """
        Converts ``RooHist`` into a ``boost-histogram`` object.
        """
        boost_histogram = uproot.extras.boost_histogram()
        bin_centers, values = self.values()
        bin_edges_low = bin_centers - self.errors(which="low", axis="x")
        bin_edges_high = bin_centers + self.errors(which="high", axis="x")
        # Boost histogram only supports symmetric errors
        if not numpy.all(numpy.isclose(bin_edges_low[1:], bin_edges_high[:-1])):
            raise ValueError("bin_edges_low[1:] != bin_edges_high[:-1]")
        errors = self.errors(which="mean", axis="y")
        variances = numpy.square(errors)

        # Now make the Boost histogram
        bin_edges = numpy.append(bin_edges_low, [bin_edges_high[-1]])
        axis = boost_histogram.axis.Variable(
            bin_edges,
            underflow=False,
            overflow=False,
        )
        axis.name = self.name
        axis.title = self.title
        hist = boost_histogram.Histogram(axis, storage=boost_histogram.storage.Weight())
        hist.name = self.name
        hist.title = self.title
        view = hist.view()
        view.value = values
        view.variance = variances
        return hist

    def to_hist(self):
        """
        Converts ``RooHist`` into a ``hist`` object.
        """
        return uproot.extras.hist().Hist(self.to_boost())

    def _to_boost_histogram_(self):
        return self.to_boost()



# BSD 3-Clause License; see https://github.com/scikit-hep/uproot4/blob/main/LICENSE

import pytest
import skhep_testdata

import uproot


def test_common_offsets():
    # this file has terrible branch alignment
    with uproot.open(skhep_testdata.data_path("uproot-hepdata-example.root")) as f:
        assert f["ntuple;1"].common_entry_offsets() == [0, 75000]

    # this file has just one branch
    with uproot.open(skhep_testdata.data_path("uproot-foriter.root")) as f:
        assert f["foriter;1"].common_entry_offsets() == [
            0,
            6,
            12,
            18,
            24,
            30,
            36,
            42,
            46,
        ]

    with uproot.open(skhep_testdata.data_path("uproot-small-dy-nooffsets.root")) as f:
        assert f["tree;1"].common_entry_offsets() == [0, 200, 400, 501]
        assert f["tree;1"].common_entry_offsets(filter_name=["Jet_pt"]) == [
            0,
            200,
            397,
            400,
            501,
        ]
        assert f["tree;1"].common_entry_offsets(filter_name=["Jet_pt", "nJet"]) == [
            0,
            200,
            400,
            501,
        ]
